from typing import Any, Dict, List, NamedTuple, Tuple, Union

from rotkehlchen.assets.asset import EthereumToken
from rotkehlchen.assets.unknown_asset import UNKNOWN_TOKEN_KEYS, UnknownEthereumToken
from rotkehlchen.constants.misc import ZERO
from rotkehlchen.serialization.deserialize import (
    deserialize_asset_amount,
    deserialize_ethereum_address,
    deserialize_ethereum_token_from_db,
    deserialize_location_from_db,
    deserialize_timestamp,
    deserialize_unknown_ethereum_token_from_db,
)
from rotkehlchen.typing import (
    AssetAmount,
    ChecksumEthAddress,
    Fee,
    Location,
    Price,
    Timestamp,
    TradePair,
    TradeType,
)

# Quick lookup for AMMTrade locations
AMMTradeLocations = {Location.UNISWAP}
AMMSwapDBTuple = (
    Tuple[
        str,  # tx_hash
        int,  # log_index
        str,  # address
        str,  # from_address
        str,  # to_address
        int,  # timestamp
        str,  # location
        int,  # is_token0_unknown
        str,  # token0_address
        str,  # token0_symbol
        str,  # token0_name
        int,  # token0_decimals
        int,  # is_token1_unknown
        str,  # token1_address
        str,  # token1_symbol
        str,  # token1_name
        int,  # token1_decimals
        str,  # amount0_in
        str,  # amount1_in
        str,  # amount0_out
        str,  # amount1_out
    ]
)


def serialize_known_or_unknown_token(
        token: Union[EthereumToken, UnknownEthereumToken],
) -> Union[str, Dict[str, Any]]:
    return (
        token.serialize()
        if isinstance(token, EthereumToken)
        else token.serialize_as_dict(keys=UNKNOWN_TOKEN_KEYS)
    )


class AMMSwap(NamedTuple):
    """Represents a single AMM swap

    For now only catering to uniswap. We will see for more when we integrate them
    """
    tx_hash: str
    log_index: int
    address: ChecksumEthAddress
    from_address: ChecksumEthAddress
    to_address: ChecksumEthAddress
    timestamp: Timestamp
    location: Location
    token0: Union[EthereumToken, UnknownEthereumToken]
    token1: Union[EthereumToken, UnknownEthereumToken]
    amount0_in: AssetAmount
    amount1_in: AssetAmount
    amount0_out: AssetAmount
    amount1_out: AssetAmount

    def __hash__(self) -> int:
        return hash(self.tx_hash + str(self.log_index))

    def __eq__(self, other: Any) -> bool:
        if other is None:
            return False

        return hash(self) == hash(other)

    @classmethod
    def deserialize_from_db(
            cls,
            trade_tuple: AMMSwapDBTuple,
    ) -> 'AMMSwap':
        """Turns a tuple read from DB into an appropriate Swap.

        May raise a DeserializationError if something is wrong with the DB data

        Trade_tuple index - Schema columns
        ----------------------------------
        0 - tx_hash
        1 - log_index
        2 - address
        3 - from_address
        4 - to_address
        5 - timestamp
        6 - location
        7 - is_token0_unknown
        8 - token0_address
        9 - token0_symbol
        10 - token0_name
        11 - token0_decimals
        12 - is_token1_unknown
        13 - token1_address
        14 - token1_symbol
        15 - token1_name
        16 - token1_decimals
        17 - amount0_in
        18 - amount1_in
        19 - amount0_out
        20 - amount1_out
        """
        address = deserialize_ethereum_address(trade_tuple[2])
        from_address = deserialize_ethereum_address(trade_tuple[3])
        to_address = deserialize_ethereum_address(trade_tuple[4])
        is_token0_unknown = trade_tuple[7]
        is_token1_unknown = trade_tuple[12]

        token0: Union[EthereumToken, UnknownEthereumToken]
        token1: Union[EthereumToken, UnknownEthereumToken]
        if is_token0_unknown:
            token0 = deserialize_unknown_ethereum_token_from_db(
                ethereum_address=trade_tuple[8],
                symbol=trade_tuple[9],
                name=trade_tuple[10],
                decimals=trade_tuple[11],
            )
        else:
            token0 = deserialize_ethereum_token_from_db(identifier=trade_tuple[9])

        if is_token1_unknown:
            token1 = deserialize_unknown_ethereum_token_from_db(
                ethereum_address=trade_tuple[13],
                symbol=trade_tuple[14],
                name=trade_tuple[15],
                decimals=trade_tuple[16],
            )
        else:
            token1 = deserialize_ethereum_token_from_db(identifier=trade_tuple[14])

        return cls(
            tx_hash=trade_tuple[0],
            log_index=trade_tuple[1],
            address=address,
            from_address=from_address,
            to_address=to_address,
            timestamp=deserialize_timestamp(trade_tuple[5]),
            location=deserialize_location_from_db(trade_tuple[6]),
            token0=token0,
            token1=token1,
            amount0_in=deserialize_asset_amount(trade_tuple[17]),
            amount1_in=deserialize_asset_amount(trade_tuple[18]),
            amount0_out=deserialize_asset_amount(trade_tuple[19]),
            amount1_out=deserialize_asset_amount(trade_tuple[20]),
        )

    def serialize(self) -> Dict[str, Any]:
        """Serialize the swap into the format seen in the frontend"""
        return {
            'tx_hash': self.tx_hash,
            'log_index': self.log_index,
            'from_address': self.from_address,
            'to_address': self.to_address,
            'location': str(self.location),
            'token0': serialize_known_or_unknown_token(self.token0),
            'token1': serialize_known_or_unknown_token(self.token1),
            'amount0_in': str(self.amount0_in),
            'amount1_in': str(self.amount1_in),
            'amount0_out': str(self.amount0_out),
            'amount1_out': str(self.amount1_out),
        }

    def to_db_tuple(self) -> AMMSwapDBTuple:
        is_token0_unknown = (
            1 if isinstance(self.token0, UnknownEthereumToken) else 0
        )
        is_token1_unknown = (
            1 if isinstance(self.token1, UnknownEthereumToken) else 0
        )
        db_tuple = (
            self.tx_hash,
            self.log_index,
            str(self.address),
            str(self.from_address),
            str(self.to_address),
            int(self.timestamp),
            self.location.serialize_for_db(),
            is_token0_unknown,
            str(self.token0.ethereum_address),
            str(self.token0.symbol),
            str(self.token0.name),
            self.token0.decimals,
            is_token1_unknown,
            str(self.token1.ethereum_address),
            str(self.token1.symbol),
            str(self.token1.name),
            self.token1.decimals,
            str(self.amount0_in),
            str(self.amount1_in),
            str(self.amount0_out),
            str(self.amount1_out),
        )
        return db_tuple  # type: ignore


class AMMTrade(NamedTuple):
    """This class aims for a better trades support than the current Trade class
    in AMMs protocols.
    """
    trade_type: TradeType
    base_asset: Union[EthereumToken, UnknownEthereumToken]
    quote_asset: Union[EthereumToken, UnknownEthereumToken]
    amount: AssetAmount
    rate: Price
    swaps: List[AMMSwap]
    # Since these trades are made up, we need an extra id for uniqueness at construction
    trade_index: int

    @property
    def tx_hash(self) -> str:
        return self.swaps[0].tx_hash

    @property
    def timestamp(self) -> Timestamp:
        return self.swaps[0].timestamp

    @property
    def location(self) -> Location:
        return self.swaps[0].location

    @property
    def address(self) -> ChecksumEthAddress:
        return self.swaps[0].address

    @property
    def identifier(self) -> str:
        """Since these trades are made up, uniqueness needs an extra id"""
        return f'{self.tx_hash}-{self.trade_index}'

    @property
    def pair(self) -> TradePair:
        return TradePair(f'{self.base_asset.symbol}_{self.quote_asset.symbol}')

    @property
    def trade_id(self) -> str:
        return self.identifier

    @property
    def fee_currency(self) -> EthereumToken:
        # always 0 DAI, super hacky -- just to satisfy the "Trade" interface for
        # use in accounting module. We need to fix these stuff ...
        return EthereumToken('DAI')

    @property
    def fee(self) -> Fee:
        return Fee(ZERO)

    def serialize(self) -> Dict[str, Any]:
        """Serialize the trade into a dict matching Trade serialization from
        `data_structures.py` and adding: address, base_asset, quote_asset,
        """
        fee_symbol = (
            self.quote_asset.identifier
            if isinstance(self.quote_asset, EthereumToken)
            else self.quote_asset.symbol
        )
        return {
            # Shared attributes with Trade
            'trade_id': self.trade_id,  # PK
            'timestamp': self.timestamp,
            'location': str(self.location),
            'pair': self.pair,
            'trade_type': str(self.trade_type),
            'amount': str(self.amount),
            'rate': str(self.rate),
            'fee': '0',  # Don't bother with a fee at the moment -- too complicated
            'fee_currency': fee_symbol,
            # AMMTrade attributes
            'address': self.address,
            'base_asset': serialize_known_or_unknown_token(self.base_asset),
            'quote_asset': serialize_known_or_unknown_token(self.quote_asset),
            'tx_hash': self.tx_hash,
            'swaps': [x.serialize() for x in self.swaps],
        }



# -*- coding: utf-8 -*-
"""
Created on Sun Dec 16 19:09:51 2018

@author: Martin
"""

import PyQt5.QtGui as qtgui
import PyQt5.QtWidgets as qtwidgets
import PyQt5.QtCore as qtcore
import PyQt5.QtChart as qtchart
import koalafolio.gui.Qcontrols as controls
import koalafolio.gui.QCharts as charts
import koalafolio.PcpCore.core as core
import koalafolio.gui.QSettings as settings
import koalafolio.gui.QStyle as style
import colorsys
import koalafolio.gui.QThreads as threads
import datetime
import koalafolio.gui.QLogger as logger
from PIL.ImageQt import ImageQt
import os
import configparser

qt = qtcore.Qt
localLogger = logger.globalLogger

class MinWheelScrollingScrollbar(qtwidgets.QScrollBar):
    def __init__(self, orientation, parent):
        super(MinWheelScrollingScrollbar, self).__init__(orientation=orientation, parent=parent)

    def wheelEvent(self, event: qtgui.QWheelEvent):
        numPixels = event.pixelDelta()
        numDegrees = event.angleDelta() / 8

        if numPixels:
            self.scrollWithPixels(numPixels)
        elif numDegrees:
            numSteps = numDegrees / 15
            self.scrollWithDegrees(numSteps.y())

        event.accept()

    def scrollWithPixels(self, numPixels):
        self.setValue(self.value() - numPixels)

    def scrollWithDegrees(self, numSteps):
        self.setValue(self.value() - self.singleStep() * numSteps)


# %% portfolio table view
class QPortfolioTableView(qtwidgets.QTreeView):
    def __init__(self, parent, *args, **kwargs):
        super(QPortfolioTableView, self).__init__(parent=parent, *args, **kwargs)

        self.setItemDelegate(QCoinTableDelegate())
        # self.setRootIsDecorated(False)

        self.setSortingEnabled(True)
        # self.setEditTriggers(qtwidgets.QAbstractItemView.NoEditTriggers)
        # self.setSelectionMode(qtwidgets.QAbstractItemView.NoSelection)

        self.visibleEditorIndex = []
        self.collapsed.connect(self.collapsedCallback)
        self.expanded.connect(self.expandedCallback)

        self.setVerticalScrollBar(MinWheelScrollingScrollbar(orientation=qtcore.Qt.Vertical, parent=self))
        self.verticalScrollBar().setSingleStep(1)
        self.verticalScrollBar().setPageStep(1)

        self.horizontalHeader().sectionResized.connect(lambda index, oldSize, newSize: self.sectionSizeChanged(index, oldSize, newSize))

    def sectionSizeChanged(self, index, oldSize, newSize):
        sourceModel = self.model().sourceModel()
        if index >= sourceModel.firstValueColumn:
            self.updateValueColumnCount(oldSize, newSize)

    def updateValueColumnCount(self, oldSize, newSize, forceUpdate=False):
        sourceModel = self.model().sourceModel()
        if forceUpdate:
            if newSize >= 220:
                sourceModel.changeValueColumnWidth(3)
            if (newSize >= 150 and newSize < 220):
                sourceModel.changeValueColumnWidth(2)
            if newSize < 150:
                sourceModel.changeValueColumnWidth(1)
        else:
            if oldSize < 220 and newSize >= 220:
                sourceModel.changeValueColumnWidth(3)
            if (oldSize < 150 or oldSize >= 220) and (newSize >= 150 and newSize < 220):
                sourceModel.changeValueColumnWidth(2)
            if oldSize >= 150 and newSize < 150:
                sourceModel.changeValueColumnWidth(1)

    def initView(self):
        self.horizontalHeader().setSectionResizeMode(qtwidgets.QHeaderView.Stretch)
        self.horizontalHeader().setSectionResizeMode(0, qtwidgets.QHeaderView.Interactive)
        self.horizontalHeader().setSectionResizeMode(1, qtwidgets.QHeaderView.Interactive)
        self.setColumnWidth(1, 160)
        self.horizontalHeader().setSectionResizeMode(2, qtwidgets.QHeaderView.Interactive)
        self.setColumnWidth(2, 100)
        # self.verticalHeader().setSectionResizeMode(qtwidgets.QHeaderView.ResizeToContents)
        # self.verticalHeader().setVisible(False)
        valueSectionSize = self.horizontalHeader().sectionSize(self.model().sourceModel().firstValueColumn)
        self.updateValueColumnCount(valueSectionSize, valueSectionSize, forceUpdate=True)

    def reset(self):
        super(QPortfolioTableView, self).reset()
        self.initView()

    def setModel(self, model):
        super(QPortfolioTableView, self).setModel(model)
        self.initView()

    def horizontalHeader(self):
        return self.header()

    def collapsedCallback(self, index):
        child = self.model().index(0, 0, index)
        child2 = self.model().index(0, 1, index)
        self.closePersistentEditor(child)
        self.closePersistentEditor(child2)

    def expandedCallback(self, index):
        child = self.model().index(0, 0, index)
        child2 = self.model().index(0, 1, index)
        self.openPersistentEditor(child)
        self.openPersistentEditor(child2)

    def drawRow(self, painter, options, index):
        if index.parent().isValid():
            firstSection = self.header().logicalIndex(0)
            # midSection = self.header().logicalIndex(2)
            lastSection = self.header().logicalIndex(self.header().count() - 1)
            left = self.header().sectionViewportPosition(firstSection)
            # mid = self.header().sectionViewportPosition(midSection) + self.header().sectionSize(midSection)
            right = self.header().sectionViewportPosition(lastSection) + self.header().sectionSize(lastSection)
            indent = 1 * self.indentation()
            left += indent

            options.rect.setX(left)
            options.rect.setWidth(right - left)

            # self.itemDelegate(index).paint(painter, options, index)
        else:
            super(QPortfolioTableView, self).drawRow(painter, options, index)
        painter.save()
        myCol = style.myStyle.getQColor('BACKGROUND_BITDARK')
        myBrush = qtgui.QBrush(myCol)
        mypen = painter.pen()
        mypen.setBrush(myBrush)
        painter.setPen(mypen)
        rect = options.rect
        if index.parent().isValid():  # draw lower line over indent
            # draw lower horz line
            x = rect.x() - self.indentation()
            painter.drawLine(x, rect.y() + rect.height(), rect.x() + rect.width(), rect.y() + rect.height())
            # draw upper horz line
            painter.drawLine(x, rect.y(), rect.x() + rect.width(), rect.y())
            # draw mid line
            mid = rect.width() * 0.3 + 20
            painter.drawLine(mid, rect.y(), mid, rect.y() + rect.height())
        else:
            # draw lower horz line
            painter.drawLine(rect.x(), rect.y() + rect.height(), rect.x() + rect.width(), rect.y() + rect.height())
        painter.restore()

# %% portfolio coin container
class QCoinContainer(qtcore.QAbstractItemModel, core.CoinList):
    coinAdded = qtcore.pyqtSignal([list])
    PriceUpdateRequest = qtcore.pyqtSignal([list])

    def __init__(self, dataPath=None, *args, **kwargs):
        super(QCoinContainer, self).__init__(*args, **kwargs)

        self.dataPath = dataPath
        self.coinDatabase = QCoinInfoDatabase(path=dataPath)
        # self.coinAdded.connect(self.saveCoins)

    def setPath(self, path):
        self.dataPath = path
        self.coinDatabase.setPath(path)

    def restoreCoins(self):
        if self.dataPath:
            # try to restore data from previous session
            coins = self.coinDatabase.getCoins()
            for coin in coins:
                if not self.hasMember(coin):
                    self.addCoin(coin)
            self.coinDatabase.updateCoinInfo(self)
        self.coinAdded.emit(self.getCoinNames())

    def saveCoins(self):
        if self.dataPath:
            if os.path.isfile(os.path.join(self.dataPath, 'Coins.csv')):
                if os.path.isfile(os.path.join(self.dataPath, 'Coins.csv.bak')):
                    try:  # delete old backup
                        os.remove(os.path.join(self.dataPath, 'Coins.csv.bak'))
                    except Exception as ex:
                        localLogger.warning('error deleting coin info backup in QCoinContainer: ' + str(ex))
                try:  # create backup
                    os.rename(os.path.join(self.dataPath, 'Coins.csv'),
                              os.path.join(self.dataPath, 'Coins.csv.bak'))
                except Exception as ex:
                    localLogger.warning('error creating coin info backup in QCoinContainer: ' + str(ex))
            self.coinDatabase.setCoinInfo(self)

    def setNotes(self, index, notes):
        self.coins[index].notes = notes
        self.saveCoins()

# %% portfolio table model
class QPortfolioTableModel(QCoinContainer):
    triggerViewReset = qtcore.pyqtSignal()

    def __init__(self, *args, **kwargs):
        super(QPortfolioTableModel, self).__init__(*args, **kwargs)

        # init keys and header
        self.valueCols = 3
        self.firstValueColumn = 3
        self.parents = []
        self.initDisplayCurrencies()

    def initDisplayCurrencies(self):
        self.keys = [*core.CoinValue().value]
        self.setColumnHeader(self.valueCols)

    def rowCount(self, parent):
        if parent.isValid():
            if not parent.parent().isValid():  # first child level
                return 1
            else:
                return 0  # only one child level
        else:  # top level
            return len(self.coins)

    def columnCount(self, parent):
        if parent.isValid():
            if not parent.parent().isValid():  # first child level
                return 2
            else:
                return 0  # only one child level
        else:  # top level
            return len(self.header)

    def parent(self, index):
        try:
            return self.parents[index.internalId()]
        except IndexError:
            return qtcore.QModelIndex()

    def index(self, row, column, parent):
        # save parent
        if parent in self.parents:
            parentId = self.parents.index(parent)
        else:
            self.parents.append(parent)
            parentId = len(self.parents)-1
        if not parent.isValid():  # top level
            return self.createIndex(row, column, parentId)
        if parent.isValid() and not parent.parent().isValid():  # first child level
            return self.createIndex(row, column, parentId)
        # only one child level
        return qtcore.QModelIndex()

    def data(self, index, role=qt.DisplayRole):
        if index.parent().isValid():  # child level
            if role == qt.DisplayRole:
                if index.column() == 0:  # coin row
                    return self.coins[index.parent().row()].notes
                if index.column() == 1:
                    return self.coins[index.parent().row()]
        else:  # top level
            if role == qt.DisplayRole:
                if index.column() == 0:  # coin row
                    return self.coins[index.row()].coinname  # return coinname
                if index.column() == 1:  # balance row
                    return self.coins[index.row()]  # return CoinBalance
                if index.column() == 2:  # profit row
                    return self.coins[index.row()].tradeMatcher.getTotalProfit()  # return profit
                if index.column() >= self.firstValueColumn:  # return CoinBalance and key
                    keys = [*core.CoinValue().value]
                    return self.coins[index.row()], keys[index.column() - self.firstValueColumn], self.valueCols
            if role == qt.DecorationRole:
                if index.column() == 0:
                    coinIcon = self.coins[index.row()].coinIcon
                    if coinIcon:
                        return coinIcon
                    else:
                        return qtcore.QVariant()
        return qtcore.QVariant()

    def headerData(self, section, orientation, role):
        if role == qt.DisplayRole:
            if orientation == qt.Horizontal:
                return self.header[section]
            elif orientation == qt.Vertical:
                return section
        if role == qt.ToolTipRole:
            if orientation == qt.Horizontal:
                if section >= 3:
                    return self.valueHeaderToolTip
        return qtcore.QVariant()

    def setColumnHeader(self, cols):
        if cols == 3:
            self.valueHeaderToolTip = 'invest value\t\tcurrent value\t\tperformance\n' + \
                                        'invest price\t\tcurrent price\t\tchange/24h'
            self.header = ['coin', 'balance', 'realized ' + settings.mySettings.reportCurrency()] + \
                          ['performance ' + key for key in self.keys]
        elif cols == 2:
            self.valueHeaderToolTip = 'current value\t\tperformance\n' + \
                                      'current price\t\tchange/24h'
            self.header = ['coin', 'balance', 'realized ' + settings.mySettings.reportCurrency()] + \
                          ['performance ' + key for key in self.keys]
        else:
            self.valueHeaderToolTip = 'current price\nchange/24h'
            self.header = ['coin', 'balance', 'realized ' + settings.mySettings.reportCurrency()] + \
                          ['price ' + key for key in self.keys]

    def changeValueColumnWidth(self, cols):
        self.valueCols = cols
        self.setColumnHeader(cols)
        self.headerDataChanged.emit(qt.Horizontal, 0, len(self.header) - 1)
        RowStartIndex = self.index(0, self.firstValueColumn, qtcore.QModelIndex())
        RowEndIndex = self.index(len(self.coins)-1, len(self.header) - 1, qtcore.QModelIndex())
        self.dataChanged.emit(RowStartIndex, RowEndIndex)

    def flags(self, index):
        if index.parent().isValid():  # child level
            return qt.ItemIsEditable | qt.ItemIsEnabled
        # top level
        return qt.ItemIsEnabled

    def setData(self, index, value, role):
        if index.parent().isValid():  # child level
            if role == qt.EditRole:
                self.setNotes(index.parent().row(), value)
        return True

    def histPricesChanged(self):
        super(QPortfolioTableModel, self).histPricesChanged()
        self.pricesUpdated()

    def histPriceUpdateFinished(self):
        super(QPortfolioTableModel, self).histPriceUpdateFinished()
        self.pricesUpdated()

    def pricesUpdated(self):
        RowStartIndex = self.index(0, 2, qtcore.QModelIndex())
        RowEndIndex = self.index(len(self.coins)-1, len(self.header) - 1, qtcore.QModelIndex())
        self.dataChanged.emit(RowStartIndex, RowEndIndex)

    def setPrices(self, prices):
        super(QPortfolioTableModel, self).setPrices(prices)
        self.pricesUpdated()

    def setPriceChartData(self, priceChartData):
        super(QPortfolioTableModel, self).setPriceChartData(priceChartData)

    # emit dataChanged when trade is updated
    def tradeChanged(self, trade):
        coin = super(QPortfolioTableModel, self).tradeChanged(trade)
        if coin:
            row = self.coins.index(coin)
            RowStartIndex = self.index(row, 0)
            RowEndIndex = self.index(row, len(self.header) - 1)
            self.dataChanged.emit(RowStartIndex, RowEndIndex)

    # emit layout changed when coin is added
    def addTrades(self, trades):
        self.beginResetModel()
        super(QPortfolioTableModel, self).addTrades(trades)
        self.endResetModel()
        #  load prices and icons of all coins (not necessary but faster)
        # todo: only pass new coins
        self.coinAdded.emit(self.getCoinNames())
        self.saveCoins()

    def triggerPriceUpdate(self):
        self.PriceUpdateRequest.emit(self.getCoinNames())

    def addCoin(self, coinname):
        newCoin = super(QPortfolioTableModel, self).addCoin(coinname)
        # coin update will be done for all coins at once since it is much faster
        # self.coinAdded.emit([newCoin.coinname])
        return newCoin

    def setIcons(self, icons):
        for coin in icons:
            try:
                self.getCoinByName(coin).coinIcon = icons[coin]
            except:
                pass
        RowStartIndex = self.index(0, 0, qtcore.QModelIndex())
        RowEndIndex = self.index(len(self.coins) - 1, 0, qtcore.QModelIndex())
        self.dataChanged.emit(RowStartIndex, RowEndIndex)

    def deleteTrades(self, trades):
        self.beginResetModel()
        super(QPortfolioTableModel, self).deleteTrades(trades)
        self.endResetModel()

    def clearCoins(self):
        self.beginResetModel()
        self.coins = []
        self.parents = []
        self.endResetModel()

    def updateDisplayCurrencies(self, keys):
        self.initDisplayCurrencies()
        self.clearCoins()
        self.restoreCoins()
        self.triggerViewReset.emit()


class QTableSortingModel(qtcore.QSortFilterProxyModel):
    def __init__(self, *args, **kwargs):
        super(QTableSortingModel, self).__init__(*args, **kwargs)

        self.sortedRow = 0
        self.sortedDir = 0

    def sort(self, row, order):
        super(QTableSortingModel, self).sort(row, order)
        self.sortedRow = row
        self.sortedDir = order

    def lessThan(self, index1, index2):
        column = index1.column()
        if column == 2:
            profit1 = index1.data()[settings.mySettings.reportCurrency()]
            profit2 = index2.data()[settings.mySettings.reportCurrency()]
            return profit1 < profit2
        if column >= self.sourceModel().firstValueColumn:
            coinBalance1, key1, cols1 = index1.data()
            coinBalance2, key2, cols2 = index2.data()
            if cols1 >= 2:
                return coinBalance1.getCurrentValue()[key1] < coinBalance2.getCurrentValue()[key2]
            else:
                return coinBalance1.getChange24h(key1) < coinBalance2.getChange24h(key2)
        return index1.data() < index2.data()

    def filterAcceptsRow(self, source_row, source_parent):
        if settings.mySettings.getGuiSetting('hideLowBalanceCoins'):
            index = self.sourceModel().index(source_row, 1, source_parent)
            data = self.sourceModel().data(index)
            if data.balance <= 0:
                return False
        if settings.mySettings.getGuiSetting('hideLowValueCoins'):
            index = self.sourceModel().index(source_row, 1, source_parent)
            data = self.sourceModel().data(index)
            if data.getCurrentValue()[settings.mySettings.reportCurrency()] <= settings.mySettings.getGuiSetting('lowValueFilterLimit(reportCurrency)'):
                return False
        return True

# %% portfolio table delegate
# class QCoinBalanceDelegate(qtwidgets.QStyledItemDelegate):
class QCoinTableDelegate(qtwidgets.QStyledItemDelegate):
    def __init__(self, *args, **kwargs):
        super(QCoinTableDelegate, self).__init__(*args, **kwargs)

        self.marginV = 4  # vertical margin
        self.marginH = 15  # horizontal margin
        self.lastCols = 0

    def paint(self, painter, option, index):
        cellStartX = option.rect.x() + self.marginH
        cellStartY = option.rect.y() + self.marginV
        cellWidth = option.rect.width() - 2 * self.marginH
        cellHeight = option.rect.height() - 2 * self.marginV
        cellStopX = cellStartX + cellWidth
        cellStopY = cellStartY + cellHeight

        painter.save()
        myCol = style.myStyle.getQColor('BACKGROUND_BITDARK')
        myBrush = qtgui.QBrush(myCol)
        mypen = painter.pen()
        mypen.setBrush(myBrush)
        painter.setPen(mypen)
        rect = option.rect
        # draw right line
        painter.drawLine(rect.x() + rect.width(), rect.y(), rect.x() + rect.width(), rect.y() + rect.height())
        painter.restore()

        contentRect = qtcore.QRect(cellStartX, cellStartY, cellWidth, cellHeight)
        painter.save()
        if not index.parent().isValid():  # top level
            if index.column() == 0:  # coin
                super(QCoinTableDelegate, self).paint(painter, option, index)
            elif index.column() == 1:  # balance
                coinBalance = index.data(qt.DisplayRole)
                balance = controls.floatToString(coinBalance.balance, 5)
                if settings.mySettings.getTaxSetting('taxfreelimit'):
                    taxFreeLimitYears = settings.mySettings.getTaxSetting('taxfreelimityears')
                    freeBuyYears = '>' + str(taxFreeLimitYears) + 'y'
                    buyAmountVal = coinBalance.tradeMatcher.getBuyAmountLeftTaxFree(taxFreeLimitYears)
                    buyAmount = controls.floatToString(buyAmountVal, 4)
                else:
                    buyAmount = controls.floatToString(coinBalance.tradeMatcher.getBuyAmount(), 4)
                    sellAmount = controls.floatToString(coinBalance.tradeMatcher.getSellAmount(), 4)
                    firstBuyLeftDate = coinBalance.tradeMatcher.getFirstBuyLeftDate()
                    if firstBuyLeftDate:
                        now = datetime.datetime.now()
                        yearDif = now.year - firstBuyLeftDate.year
                        monthDif = now.month - firstBuyLeftDate.month
                        dayDif = now.day - firstBuyLeftDate.day
                        if monthDif < 0 or (monthDif == 0 and dayDif < 0):
                            yearDif -= 1
                        freeBuyYears = '>' + str(yearDif) + 'y'

                defaultFont = painter.font()
                # paint balance text
                newFont = painter.font()
                newFont.setPixelSize(15)
                painter.setFont(newFont)
                painter.drawText(contentRect, qt.AlignHCenter | qt.AlignTop, balance)
                painter.setFont(defaultFont)

                if settings.mySettings.getTaxSetting('taxfreelimit'):
                    if buyAmountVal > 0:
                        # paint buy that can be sold tax free
                        buyColor = qtgui.QColor(*settings.mySettings.getColor('POSITIV'))
                        buyBrush = qtgui.QBrush(buyColor)
                        buyPen = painter.pen()
                        buyPen.setBrush(buyBrush)
                        painter.setPen(buyPen)
                        painter.drawText(contentRect, qt.AlignLeft | qt.AlignBottom, buyAmount)
                        painter.drawText(contentRect, qt.AlignLeft | qt.AlignTop, freeBuyYears)
                else:
                    # paint buy
                    buyColor = qtgui.QColor(*settings.mySettings.getColor('POSITIV'))
                    buyBrush = qtgui.QBrush(buyColor)
                    buyPen = painter.pen()
                    buyPen.setBrush(buyBrush)
                    painter.setPen(buyPen)
                    painter.drawText(contentRect, qt.AlignLeft | qt.AlignBottom, buyAmount)
                    if firstBuyLeftDate:
                        painter.drawText(contentRect, qt.AlignLeft | qt.AlignTop, freeBuyYears)

                    # paint sell
                    sellColor = qtgui.QColor(*settings.mySettings.getColor('NEGATIV'))
                    sellBrush = qtgui.QBrush(sellColor)
                    sellPen = painter.pen()
                    sellPen.setBrush(sellBrush)
                    painter.setPen(sellPen)
                    painter.drawText(contentRect, qt.AlignRight | qt.AlignBottom, sellAmount)

            elif index.column() == 2:
                profit = index.data(qt.DisplayRole)
                key = settings.mySettings.reportCurrency()
                # for key in profit:
                # draw profit
                if profit[key] >= 0:
                    drawColor = qtgui.QColor(*settings.mySettings.getColor('POSITIV'))
                else:
                    drawColor = qtgui.QColor(*settings.mySettings.getColor('NEGATIV'))
                pen = painter.pen()
                pen.setBrush(qtgui.QBrush(drawColor))
                painter.setPen(pen)
                defaultFont = painter.font()
                newFont = painter.font()
                newFont.setPixelSize(14)
                painter.setFont(newFont)
                profitString = controls.floatToString(profit[key], 4)
                painter.drawText(contentRect, qt.AlignHCenter | qt.AlignVCenter, profitString)

            elif index.column() >= index.model().sourceModel().firstValueColumn:
                try:  # catch key error if currency is not in database. Could happen during currency switch
                    coinBalance, key, cols = index.data(qt.DisplayRole)
                    boughtValue = coinBalance.initialValue[key]
                    boughtPrice = coinBalance.getInitialPrice()[key]
                    currentValue = coinBalance.getCurrentValue()[key]
                    currentPrice = coinBalance.getCurrentPrice()[key]
                    boughtValueStr = (controls.floatToString(boughtValue, 4))
                    boughtPriceStr = (controls.floatToString(boughtPrice, 4) + '/p')
                    currentValueStr = (controls.floatToString(currentValue, 4))
                    currentPriceStr = (controls.floatToString(currentPrice, 4) + '/p')
                    if boughtPrice == 0:
                        gain = 0
                    else:
                        gain = (currentPrice / boughtPrice - 1) * 100
                    gainStr = ("%.2f%%" % (gain))
                    gainDay = coinBalance.getChange24h(key)
                    gainDayStr = ("%.2f%%/24h" % (gainDay))


                    positivColor = style.myStyle.getQColor('POSITIV')
                    negativColor = style.myStyle.getQColor('NEGATIV')
                    neutralColor = style.myStyle.getQColor('TEXT_NORMAL')

                    def drawText(alignHorz, alignVert, fontSize, color, text):
                        newFont = painter.font()
                        newFont.setPixelSize(fontSize)
                        painter.setFont(newFont)
                        pen = painter.pen()
                        pen.setColor(color)
                        painter.setPen(pen)
                        painter.drawText(contentRect, alignHorz | alignVert, text)

                    if cols == 3:  # draw all
                        drawText(qt.AlignLeft, qt.AlignTop, 14, neutralColor, boughtValueStr)
                        drawText(qt.AlignLeft, qt.AlignBottom, 12, neutralColor, boughtPriceStr)
                        if gain > 0:
                            drawText(qt.AlignHCenter, qt.AlignTop, 14, positivColor, currentValueStr)
                            drawText(qt.AlignHCenter, qt.AlignBottom, 12, positivColor, currentPriceStr)
                            drawText(qt.AlignRight, qt.AlignTop, 14, positivColor, gainStr)
                        elif gain < 0:
                            drawText(qt.AlignHCenter, qt.AlignTop, 14, negativColor, currentValueStr)
                            drawText(qt.AlignHCenter, qt.AlignBottom, 12, negativColor, currentPriceStr)
                            drawText(qt.AlignRight, qt.AlignTop, 14, negativColor, gainStr)
                        else:
                            drawText(qt.AlignHCenter, qt.AlignTop, 14, neutralColor, currentValueStr)
                            drawText(qt.AlignHCenter, qt.AlignBottom, 12, neutralColor, currentPriceStr)
                            drawText(qt.AlignRight, qt.AlignTop, 14, neutralColor, gainStr)
                        if gainDay >= 0:
                            drawText(qt.AlignRight, qt.AlignBottom, 12, positivColor, gainDayStr)
                        else:
                            drawText(qt.AlignRight, qt.AlignBottom, 12, negativColor, gainDayStr)
                    elif cols == 2:  # skip current value and price
                        if gain > 0:
                            drawText(qt.AlignLeft, qt.AlignTop, 14, positivColor, currentValueStr)
                            drawText(qt.AlignLeft, qt.AlignBottom, 12, positivColor, currentPriceStr)
                            drawText(qt.AlignRight, qt.AlignTop, 14, positivColor, gainStr)
                        elif gain < 0:
                            drawText(qt.AlignLeft, qt.AlignTop, 14, negativColor, currentValueStr)
                            drawText(qt.AlignLeft, qt.AlignBottom, 12, negativColor, currentPriceStr)
                            drawText(qt.AlignRight, qt.AlignTop, 14, negativColor, gainStr)
                        else:
                            drawText(qt.AlignLeft, qt.AlignTop, 14, neutralColor, currentValueStr)
                            drawText(qt.AlignLeft, qt.AlignBottom, 12, neutralColor, currentPriceStr)
                            drawText(qt.AlignRight, qt.AlignTop, 14, neutralColor, gainStr)
                        if gainDay >= 0:
                            drawText(qt.AlignRight, qt.AlignBottom, 12, positivColor, gainDayStr)
                        else:
                            drawText(qt.AlignRight, qt.AlignBottom, 12, negativColor, gainDayStr)
                    else:  # draw value and daygain
                        if gainDay >= 0:
                            drawText(qt.AlignHCenter, qt.AlignTop, 14, positivColor, currentPriceStr)
                            drawText(qt.AlignHCenter, qt.AlignBottom, 12, positivColor, gainDayStr)
                        else:
                            drawText(qt.AlignHCenter, qt.AlignTop, 14, negativColor, currentPriceStr)
                            drawText(qt.AlignHCenter, qt.AlignBottom, 12, negativColor, gainDayStr)
                except KeyError as ex:
                    localLogger.warning("currency is missing in coindata: " + str(ex))
        # else:  # child level
        #     if index.column() == 1:  # trade hist chart
        #         # draw chart
        painter.restore()

    def createEditor(self, parent, option, index):
        if index.parent().isValid():  # child level
            if index.column() == 0:
                if int(index.flags()) & qt.ItemIsEditable:
                    textedit = qtwidgets.QTextEdit(parent)
                    textedit.setPlaceholderText('notes')
                    self.updateEditorGeometry(textedit, option, index)
                    return textedit
            if index.column() == 1:
                timelinechart = charts.BuyTimelineChartCont(parent)
                return timelinechart
        return None

    def setEditorData(self, editor, index):
        if index.parent().isValid():
            if index.column() == 0:
                editor.setText(index.data())
                return
            if index.column() == 1:
                data = index.data()
                # draw buys left
                dates = []
                vals = []
                oldSum = 0
                # draw buy prices
                priceDates = []
                prices = []
                for trade in data.tradeMatcher.buysLeft:
                    date = datetime.datetime.combine(trade.date, datetime.time(0, 0, 0, 0))
                    # buys left
                    dates.append(date)
                    vals.append(oldSum)
                    dates.append(date)
                    vals.append(oldSum + trade.amount)
                    oldSum = vals[-1]
                    # prices
                    priceDates.append(date)
                    prices.append(trade.getPrice()[settings.mySettings.reportCurrency()])
                if dates:
                    # buys left
                    dates.append(datetime.datetime.now())
                    vals.append(vals[-1])
                    # prices
                    if not data.priceChartData: # if not price chart data available add at least current value
                        priceDates.append(datetime.datetime.now())
                        prices.append(data.getCurrentPrice()[settings.mySettings.reportCurrency()])
                    # draw taxlimit
                    if settings.mySettings.getTaxSetting("taxfreelimit"):
                        limitDate = datetime.datetime.now().replace(year=datetime.datetime.now().year -
                                                                         settings.mySettings.getTaxSetting("taxfreelimityears"))
                        firstDate = datetime.datetime.combine(data.tradeMatcher.buysLeft[0].date, datetime.time(0, 0, 0, 0))
                        limitDates = [limitDate, limitDate]
                        limitAmount = data.tradeMatcher.getBuyAmountLeftTaxFree(settings.mySettings.getTaxSetting("taxfreelimityears"))
                        limitVals = [limitAmount, 0]
                    # addData(self, dates, vals, qColor, name, lineWidth=1, chartType="line", axis='balance', updateAxis=True)
                    editor.addData(dates, vals, style.myStyle.getQColor('PRIMARY_MIDLIGHT'), "buys", lineWidth=3)
                    if settings.mySettings.getTaxSetting("taxfreelimit"):
                        editor.addData(limitDates, limitVals, style.myStyle.getQColor('POSITIV'), "taxfree", lineWidth=3, updateAxis=False)
                    editor.addData(priceDates, prices, style.myStyle.getQColor('SECONDARY'),
                                   "buyPrice " + settings.mySettings.reportCurrency(), lineWidth=3,
                                   chartType="scatter", axis='price')
                    if data.priceChartData:
                        dates, prices = map(list, zip(*data.priceChartData))
                        editor.addData(dates, prices, style.myStyle.getQColor('SECONDARY_MIDLIGHT'),
                                       "price " + settings.mySettings.reportCurrency(), lineWidth=1,
                                       chartType="line", axis='price')
                return

    def updateEditorGeometry(self, editor, option, index):
        if index.parent().isValid():  # child level

            if index.column() == 0:
                size = self.sizeHint(option, index)
                rect = qtcore.QRect(editor.parent().x(), option.rect.y() + 5, editor.parent().width() * 0.3,
                                    size.height() - 10)
                editor.setGeometry(rect)
                return
            if index.column() == 1:
                size = self.sizeHint(option, index)
                rect = qtcore.QRect(editor.parent().x() + editor.parent().width() * 0.3 + 20, option.rect.y() + 5, editor.parent().width() * 0.7 - 20,
                                    size.height() - 10)
                editor.setGeometry(rect)
                return
        else:
            raise TypeError

    def setModelData(self, editor, model, index):
        if index.parent().isValid():
            if index.column() == 0:
                model.setData(index, editor.toPlainText(), qt.EditRole)

    def sizeHint(self, option, index):
        if index.parent().isValid():  # child level
            size = super(QCoinTableDelegate, self).sizeHint(option, index)
            return qtcore.QSize(size.width(), 200)
        else:
            if index.column() == 0:
                return qtcore.QSize(50, 40)
            elif index.column() == 1:
                return qtcore.QSize(100, 40)
            elif index.column() >= 2:
                return qtcore.QSize(200, 40)
        return qtcore.QSize()


class QArrowPainterPath(qtgui.QPainterPath):
    def __init__(self, startX, startY, width, height, *args, **kwargs):
        super(QArrowPainterPath, self).__init__(*args, **kwargs)

        self.startX = startX
        self.startY = startY

        #        self.height = 6 * size
        #        self.width = 9 * size
        self.height = height
        self.width = width

        self.stopX = self.startX + self.width
        self.stopY = self.startY + self.height

        self.moveTo(startX, startY)
        self.lineTo(startX + self.width / 9 * 5, startY)
        self.lineTo(startX + self.width / 9 * 5, startY - self.height / 6 * 2)
        self.lineTo(startX + self.width, startY + self.height / 6 * 2)
        self.lineTo(startX + self.width / 9 * 5, startY + self.height)
        self.lineTo(startX + self.width / 9 * 5, startY + self.height / 6 * 4)
        self.lineTo(startX, startY + self.height / 6 * 4)
        self.lineTo(startX, startY)
        self.closeSubpath()


# %% portfolio overview
class PortfolioOverview(qtwidgets.QWidget):
    expandTable = qtcore.pyqtSignal()
    collapseTable = qtcore.pyqtSignal()
    filterSoldCoins = qtcore.pyqtSignal()

    def __init__(self, controller, height=200, *args, **kwargs):
        super(PortfolioOverview, self).__init__(*args, **kwargs)

        self.controller = controller
        self.styleHandler = self.controller.styleSheetHandler
        self.height = height
        self.setFixedHeight(self.height)
        self.setContentsMargins(0, 0, 0, 0)
        self.horzLayout = qtwidgets.QHBoxLayout(self)
        self.horzLayout.setContentsMargins(0, 0, 0, 0)

        # profit table
        self.profitTable = qtwidgets.QTableWidget()
        self.profitTable.setSelectionMode(qtwidgets.QAbstractItemView.NoSelection)
        self.profitTable.setColumnCount(4)
        self.profitTable.setHorizontalHeaderLabels(["profit", "tax profit", "fees", "fiat profit"])
        self.profitTable.horizontalHeaderItem(0).setToolTip("profit per year")
        self.profitTable.horizontalHeaderItem(1).setToolTip("profit relevant for tax calculation")
        self.profitTable.horizontalHeaderItem(2).setToolTip("fees per year")
        self.profitTable.horizontalHeaderItem(3).setToolTip("fiat profit per year")
        self.profitTable.horizontalHeader().setVisible(True)
        self.profitTable.setSizeAdjustPolicy(qtwidgets.QAbstractScrollArea.AdjustToContents)

        # tax value chart
        self.currentValueChart = charts.LabeledDonatChart(self.height, self.height, 3,
                                                          'crypto performance')
        self.currentValueChart.setHeadingToolTip('this chart shows\nthe performance of\n' +
                                                 'your portfolio\nrelative to the\ncurrent invest\n' +
                                                 '(profit from crypto\ntrades is reinvested)')
        self.donutSliceInvested = self.currentValueChart.addSlice('invested', 1, -1, False)
        self.donutSlicePerformance = self.currentValueChart.addSlice('performance', 0.5, -1, False)
        # fiat value chart
        self.currentFiatValueChart = charts.LabeledDonatChart(self.height, self.height, 3,
                                                              'fiat performance')
        self.currentFiatValueChart.setHeadingToolTip('this chart shows\nthe performance of\nyour portfolio\n' +
                                                     'relative to the\ninitial fiat invest')
        self.sliceFiatInvested = self.currentFiatValueChart.addSlice('fiat invest', 1, -1, False)
        self.sliceCoinValue = self.currentFiatValueChart.addSlice('coin value', 0.5, -1, False)
        self.sliceFiatReturn = self.currentFiatValueChart.addSlice('fiat return', 0.5, -1, False)

        self.perfChartCont = charts.ChartCont(self)
        self.controller.startRefresh.connect(self.perfChartCont.clearButtonStyle)
        self.controller.endRefresh.connect(self.perfChartCont.setButtonStyle)
        self.perfChartCont.addChart(self.currentValueChart)
        self.perfChartCont.addChart(self.currentFiatValueChart)
        self.perfChartCont.setChartIndex(settings.mySettings.getGuiSetting('performanceChartIndex'))
        self.horzLayout.addWidget(self.perfChartCont)

        # table controls
        self.expandAllButton = qtwidgets.QPushButton("expand", self)
        self.expandAllButton.clicked.connect(self.expandTable)
        self.collapseAllButton = qtwidgets.QPushButton("collapse", self)
        self.collapseAllButton.clicked.connect(self.collapseTable)
        # self.filterSoldCoinsBox = qtwidgets.QCheckBox(self)

        self.controlsLayout = qtwidgets.QHBoxLayout()
        self.controlsLayout.addStretch()
        self.controlsLayout.addWidget(self.expandAllButton)
        self.controlsLayout.addWidget(self.collapseAllButton)
        # self.controlsLayout.addWidget(self.filterSoldCoinsBox)
        self.controlsLayout.addStretch()

        self.centerVertLayout = qtwidgets.QVBoxLayout()
        self.centerVertLayout.addWidget(self.profitTable)
        self.centerVertLayout.addStretch()
        self.centerVertLayout.addLayout(self.controlsLayout)

        self.horzLayout.addStretch()
        self.horzLayout.addLayout(self.centerVertLayout)
        self.horzLayout.addStretch()

        # pie chart
        numLabel = len(settings.mySettings.displayCurrencies())
        self.portfolioChart = charts.LabeledDonatChart(self.height + 30, self.height, numLabel, parent=self)
        self.controller.startRefresh.connect(self.portfolioChart.chartView.clearStyleSheet)
        # self.controller.endRefresh.connect(self.portfolioChart.chartView.setColor)
        self.horzLayout.addWidget(self.portfolioChart)

        self.refresh()

    def refresh(self):
        self.negColor = qtgui.QColor(*settings.mySettings.getColor('NEGATIV'))
        self.posColor = qtgui.QColor(*settings.mySettings.getColor('POSITIV'))
        self.neutrColor = qtgui.QColor(*settings.mySettings.getColor('TEXT_NORMAL'))

    def setModel(self, model):
        self.model = model
        self.model.modelReset.connect(self.coinTableChangedSlot)
        self.model.dataChanged.connect(self.coinTableChangedSlot)
        self.coinTableChangedSlot()

    def coinTableChangedSlot(self):
        # update new values

        # todo: display all displayCurrencies
        taxCoinName = settings.mySettings.reportCurrency()
        numberOfDecimals = 4

        # initialize local vars
        # total invested fiat
        totalInvestFiat = core.CoinValue()
        # total returned fiat
        totalReturnFiat = core.CoinValue()
        # fiat performance
        # fiatPerformance = core.CoinValue()

        # invested value of current holdings
        currentInvestNoFiat = core.CoinValue()
        # current value of current holdings (hypothetical)
        hypotheticalCoinValueNoFiat = core.CoinValue()
        # current performance of current holdings (hypothetical)
        # hypotheticalPerformanceNoFiat = core.CoinValue()

        # realized profit (relevant for tax)
        realizedProfit = core.CoinValue()
        # unrealized profit (would be relevant for tax if realized)
        # unrealizedProfit = core.CoinValue()
        # paid fees
        paidFees = core.CoinValue()

        # testing
        # currentInvestAll = core.CoinValue()
        # hypotheticalValueAll = core.CoinValue()
        # realizedProfitAll = core.CoinValue()
        if self.controller.tradeList.isEmpty():
            startYear = datetime.datetime.now().year
        else:
            startYear = min([trade.date for trade in self.controller.tradeList]).year
        stopYear = datetime.datetime.now().year
        realizedProfitPerYear = {}
        paidFeesPerYear = {}
        realizedProfitPerYearCoinList = {}
        fiatPerYear = {}
        taxProfitPerYear = {}

        for year in range(startYear, stopYear+1):
            realizedProfitPerYear[str(year)] = core.CoinValue()
            paidFeesPerYear[str(year)] = core.CoinValue()
            fiatPerYear[str(year)] = core.CoinValue()
            taxProfitPerYear[str(year)] = core.CoinValue()

        # calculate all needed values
        for coin in self.model:
            if coin.isFiat():  # calculate invested and returned fiat
                for trade in coin.trades:
                    if trade.amount < 0:
                        totalInvestFiat.add(trade.getValue().mult(-1))
                    else:
                        totalReturnFiat.add(trade.getValue())
                    # fiat invest/ return per year
                    for year in range(startYear, stopYear + 1):
                        startDate = datetime.date(year=year, month=1, day=1)
                        endDate = datetime.date(year=year, month=12, day=31)
                        if trade.date.date() >= startDate and trade.date.date() <= endDate:
                            fiatPerYear[str(year)].add(trade.getValue())

            else:  # calculate value of portfolio
                currentInvestNoFiat.add(coin.initialValue)
                hypotheticalCoinValueNoFiat.add(coin.getCurrentValue())
                realizedProfit.add(coin.tradeMatcher.getTotalProfit())
            # calc fees per year
            for trade in coin.trades:
                if trade.tradeType == "fee":
                    paidFees.add(trade.getValue())
                    for year in range(startYear, stopYear + 1):
                        startDate = datetime.date(year=year, month=1, day=1)
                        endDate = datetime.date(year=year, month=12, day=31)
                        if trade.date.date() >= startDate and trade.date.date() <= endDate:
                            paidFeesPerYear[str(year)].add(trade.getValue())

            realizedProfitPerYearCoinList[coin.coinname] = (coin.tradeMatcher.getTimeDeltaProfit(
                    datetime.date(year=2017, month=1, day=1), datetime.date(year=2019, month=12, day=31),
                    taxFreeTimeDelta=settings.mySettings.getTaxSetting('taxfreelimityears'))[taxCoinName])
            for year in range(startYear, stopYear + 1):
                startDate = datetime.date(year=year, month=1, day=1)
                endDate = datetime.date(year=year, month=12, day=31)
                taxProfitPerYear[str(year)].add(coin.tradeMatcher.getTimeDeltaProfit(startDate, endDate,
                    taxFreeTimeDelta=settings.mySettings.getTaxSetting('taxfreelimityears')))
                realizedProfitPerYear[str(year)].add(coin.tradeMatcher.getTimeDeltaProfit(startDate, endDate,
                    taxFreeTimeDelta = -1))
            # fiat and coins
            # currentInvestAll.add(coin.initialValue)
            # hypotheticalValueAll.add(coin.getCurrentValue())
            # realizedProfitAll.add(coin.tradeMatcher.getTotalProfit())

        realizedProfitPerYearCoinListSum = sum([realizedProfitPerYearCoinList[key] for key in realizedProfitPerYearCoinList])
        fiatPerformance = (totalReturnFiat-totalInvestFiat).div(totalInvestFiat).mult(100)
        hypotheticalPerformanceNoFiat = (hypotheticalCoinValueNoFiat.div(currentInvestNoFiat)
                                         - core.CoinValue().setValue(1)).mult(100)
        unrealizedProfit = hypotheticalCoinValueNoFiat - currentInvestNoFiat

        def setLabelColor(label, isPositiv):
            if isPositiv:
                label.setBodyColor(self.posColor.name())
            else:
                label.setBodyColor(self.negColor.name())

        # tax value chart
        self.currentValueChart.chartView.setText(
            [controls.floatToString(currentInvestNoFiat[taxCoinName], numberOfDecimals) + ' ' + taxCoinName,
            controls.floatToString(hypotheticalCoinValueNoFiat[taxCoinName], numberOfDecimals) + ' ' + taxCoinName,
            "%.2f%%" % hypotheticalPerformanceNoFiat[taxCoinName]])

        self.currentValueChart.setLabelToolTip(['current invest', 'current value', 'performance'])

        if unrealizedProfit[taxCoinName] < 0:
            self.donutSliceInvested.setValue(currentInvestNoFiat[taxCoinName]+unrealizedProfit[taxCoinName])
            self.donutSliceInvested.setColor(self.neutrColor)
            self.donutSlicePerformance.setValue(-unrealizedProfit[taxCoinName])
            self.donutSlicePerformance.setColor(self.negColor)
            # self.donutSlicePerformance.setLabelColor(self.negColor)
            self.currentValueChart.chartView.setColor([self.neutrColor, self.negColor, self.negColor])
        else:
            self.donutSliceInvested.setValue(currentInvestNoFiat[taxCoinName])
            self.donutSliceInvested.setColor(self.neutrColor)
            self.donutSlicePerformance.setValue(unrealizedProfit[taxCoinName])
            self.donutSlicePerformance.setColor(self.posColor)
            # self.donutSlicePerformance.setLabelColor(self.posColor)
            self.currentValueChart.chartView.setColor([self.neutrColor, self.posColor, self.posColor])

        # fiat value chart
        self.currentFiatValueChart.chartView.setText(
            [controls.floatToString(totalInvestFiat[taxCoinName], numberOfDecimals) + ' ' + taxCoinName,
             controls.floatToString(hypotheticalCoinValueNoFiat[taxCoinName], numberOfDecimals) + ' ' + taxCoinName,
             controls.floatToString(totalReturnFiat[taxCoinName], numberOfDecimals) + ' ' + taxCoinName], qt.AlignCenter)

        self.currentFiatValueChart.setLabelToolTip(['fiat invest', 'current value', 'fiat return'])

        self.sliceFiatInvested.setValue(totalInvestFiat[taxCoinName])
        self.sliceFiatInvested.setColor(self.neutrColor)
        self.sliceFiatReturn.setValue(totalReturnFiat[taxCoinName])
        self.sliceFiatReturn.setColor(self.styleHandler.getQColor('PRIMARY'))
        self.sliceCoinValue.setValue(hypotheticalCoinValueNoFiat[taxCoinName])

        if (hypotheticalCoinValueNoFiat[taxCoinName] + totalReturnFiat[taxCoinName]) \
                < totalInvestFiat[taxCoinName]:
            self.sliceCoinValue.setColor(self.negColor)
            self.currentFiatValueChart.chartView.setColor([self.neutrColor, self.negColor,
                                                       self.styleHandler.getQColor('PRIMARY')])
        else:
            self.sliceCoinValue.setColor(self.posColor)
            self.currentFiatValueChart.chartView.setColor([self.neutrColor, self.posColor,
                                                       self.styleHandler.getQColor('PRIMARY')])

        # profit table
        years = []
        for year in realizedProfitPerYear:
            years.append(year)
        self.profitTable.setRowCount(len(years))
        self.profitTable.setVerticalHeaderLabels(years)
        for year, row in zip(realizedProfitPerYear, range(len(realizedProfitPerYear))):
            self.profitTable.setItem(row, 0, qtwidgets.QTableWidgetItem(
                controls.floatToString(realizedProfitPerYear[year][taxCoinName], 5) + ' ' + taxCoinName))
            self.profitTable.setItem(row, 1, qtwidgets.QTableWidgetItem(
                controls.floatToString(taxProfitPerYear[year][taxCoinName], 5) + ' ' + taxCoinName))
            self.profitTable.setItem(row, 2, qtwidgets.QTableWidgetItem(
                controls.floatToString(paidFeesPerYear[year][taxCoinName], 5) + ' ' + taxCoinName))
            self.profitTable.setItem(row, 3, qtwidgets.QTableWidgetItem(
                controls.floatToString(fiatPerYear[year][taxCoinName], 5) + ' ' + taxCoinName))


        # pie chart
        pieSeries = qtchart.QPieSeries()
        sortedModelIndex = sorted(range(len(self.model)),
                                  key=lambda x: self.model.coins[x].getCurrentValue()[taxCoinName], reverse=True)
        otherssum = core.CoinValue()
        try:
            topvalue = self.model.coins[sortedModelIndex[0]].getCurrentValue()[taxCoinName]
        except IndexError:
            topvalue = 0
        for index in sortedModelIndex:
            coin = self.model.coins[index]
            if coin.getCurrentValue()[taxCoinName] > topvalue/40 and \
                    coin.getCurrentValue()[taxCoinName] > abs(hypotheticalCoinValueNoFiat[taxCoinName]/75):
                pieSeries.append(coin.coinname, coin.getCurrentValue()[taxCoinName])
            elif coin.getCurrentValue()[taxCoinName] > 0:
                otherssum.add(coin.getCurrentValue())
        if otherssum[taxCoinName] > abs(hypotheticalCoinValueNoFiat[taxCoinName]/100):
            slice = pieSeries.append("others", otherssum[taxCoinName])
            slice.setLabelVisible()

        # if len(pieSeries.slices()) > 5:
        #     for slice in pieSeries.slices()[0:5]:
        #         if slice.value() > hypotheticalCoinValueNoFiat[taxCoinName]/20:
        #             slice.setLabelVisible()
        # else:
        for slice in pieSeries.slices():
            if slice.value() > abs(hypotheticalCoinValueNoFiat[taxCoinName]/20):
                slice.setLabelVisible()

        color = [255, 75, 225]
        for slice in pieSeries.slices():
            color = style.nextColor(color, 55)
            slice.setBrush(qtgui.QColor(*tuple(color)))
            slice.setLabelColor(qtgui.QColor(*tuple(color)))
            slice.setLabelPosition(qtchart.QPieSlice.LabelOutside)

        pieSeries.setHoleSize(0.6)
        self.portfolioChart.setSeries(pieSeries)
        portfolioChartLabels = []
        for coin in hypotheticalCoinValueNoFiat:
            portfolioChartLabels.append(controls.floatToString(hypotheticalCoinValueNoFiat[coin],
                                                            numberOfDecimals) + ' ' + coin)
        self.portfolioChart.chartView.setText(portfolioChartLabels, qt.AlignCenter)
        self.portfolioChart.chartView.setColor(self.neutrColor, False)


    def displayCurrenciesChangedSlot(self):
        # todo: update ui layout and trigger value update
        self.portfolioChart.chartView.setLabelCount(len(settings.mySettings.displayCurrencies()))


# coin info file
class QCoinInfoDatabase(configparser.ConfigParser):
    def __init__(self, path=None, *args, **kwargs):
        super(QCoinInfoDatabase, self).__init__(*args, **kwargs)

        self.filePath = None
        if path:
            self.setPath(path)


    def setPath(self, path):
        self.filePath = os.path.join(path, 'coininfo.txt')
        # init settings
        if not os.path.isfile(self.filePath):
            self.saveCoins()
        else:
            self.readCoins()
            self.saveCoins()
        return self

    def saveCoins(self):
        try:
            with open(self.filePath, 'w') as coinfile:
                self.write(coinfile)
            logger.globalLogger.info('coins saved')
            return True
        except Exception as ex:
            logger.globalLogger.error('error in saveCoins: ' + str(ex))
            return False

    def readCoins(self):
        try:
            self.read(self.filePath)
            logger.globalLogger.info('coin info loaded')
        except Exception as ex:
            logger.globalLogger.error('coin info can not be loaded: ' + str(ex))

    def setCoinInfo(self, coinList):
        for coin in coinList:
            if not coin.coinname in self:
                self[coin.coinname] = {}
            self[coin.coinname]['notes'] = coin.notes
        self.saveCoins()

    def updateCoinInfo(self, coinList):
        for coin in coinList:
            try:
                coin.notes = self[coin.coinname]['notes']
            except KeyError:
                coin.notes = ""

    def getCoins(self):
        coins = []
        for key in self:
            if key != 'DEFAULT':
                coins.append(key)
        return coins



from enum import Enum


class NodeName(Enum):
    OWN = 0
    ETHERSCAN = 1
    MYCRYPTO = 2
    BLOCKSCOUT = 3
    AVADO_POOL = 4

    def __str__(self) -> str:
        if self == NodeName.OWN:
            return 'own node'
        elif self == NodeName.ETHERSCAN:
            return 'etherscan'
        elif self == NodeName.MYCRYPTO:
            return 'mycrypto'
        elif self == NodeName.BLOCKSCOUT:
            return 'blockscout'
        elif self == NodeName.AVADO_POOL:
            return 'avado pool'

        raise RuntimeError(f'Corrupt value {self} for NodeName -- Should never happen')

    def endpoint(self, own_rpc_endpoint: str) -> str:
        if self == NodeName.OWN:
            return own_rpc_endpoint
        elif self == NodeName.ETHERSCAN:
            raise TypeError('Called endpoint for etherscan')
        elif self == NodeName.MYCRYPTO:
            return 'https://api.mycryptoapi.com/eth'
        elif self == NodeName.BLOCKSCOUT:
            return 'https://mainnet-nethermind.blockscout.com/'
        elif self == NodeName.AVADO_POOL:
            return 'https://mainnet.eth.cloud.ava.do/'

        raise RuntimeError(f'Corrupt value {self} for NodeName -- Should never happen')



import logging
from collections import deque
from typing import Deque, List

logger = logging.getLogger(__name__)


class MessagesAggregator():
    """
    This class is passed around where needed and aggreggates messages for the user
    """

    def __init__(self) -> None:
        self.warnings: Deque = deque()
        self.errors: Deque = deque()

    def add_warning(self, msg: str) -> None:
        logger.warning(msg)
        self.warnings.appendleft(msg)

    def consume_warnings(self) -> List[str]:
        result = []
        while len(self.warnings) != 0:
            result.append(self.warnings.pop())
        return result

    def add_error(self, msg: str) -> None:
        logger.error(msg)
        self.errors.appendleft(msg)

    def consume_errors(self) -> List[str]:
        result = []
        while len(self.errors) != 0:
            result.append(self.errors.pop())
        return result



import logging
import os
import sys
from json.decoder import JSONDecodeError
from pathlib import Path
from typing import Any, Dict, List, Optional

import gevent
import requests

from rotkehlchen.errors import RemoteError
from rotkehlchen.logging import RotkehlchenLogsAdapter
from rotkehlchen.utils.misc import ts_now
from rotkehlchen.utils.serialization import rlk_jsondumps, rlk_jsonloads_dict

logger = logging.getLogger(__name__)
log = RotkehlchenLogsAdapter(logger)

INITIAL_BACKOFF = 5


KNOWN_TO_MISS_FROM_CMC = (
    'VEN',
    '1CR',
    'DAO',
    'KFEE',
    'BTXCRD',
    'AC',
    'ACH',
    'ADADOWN',
    'ADAUP',
    'BTCDOWN',
    'BTCUP',
    'ETHDOWN',
    'ETHUP',
    'LINKDOWN',
    'LINKUP',
    'ADN',
    'AERO',
    'AM',
    'AIR-2',
    'AIR',
    'APH-2',
    'ARCH',
    # Missing from CMC, is in cryptocompare: https://www.cryptocompare.com/coins/bidr
    'BIDR',
    # Missing from API but is in website: https://coinmarketcap.com/currencies/bitcoindark/
    'BTCD',
    # Missing from API but is in website: https://coinmarketcap.com/currencies/cachecoin/
    'CACH',
    # Missing from API is in website https://coinmarketcap.com/currencies/caix/
    'CAIX',
    # Missing from API is in website https://coinmarketcap.com/currencies/cannacoin/
    'CCN-2',
    # Missing from API is in website https://coinmarketcap.com/currencies/cryptographic-anomaly/
    'CGA',
    # Missing from API, is in website https://coinmarketcap.com/currencies/cinni/
    'CINNI',
    # Missing from API, is in website https://coinmarketcap.com/currencies/concealcoin/
    'CNL',
    # Missing from API, is in website https://coinmarketcap.com/currencies/coinomat/
    'CNMT',
    # Missing from API, is in website https://coinmarketcap.com/currencies/communitycoin/
    # Missing from CMC, is in cryptocompare: https://www.cryptocompare.com/coins/cntm
    'CNTM',
    'COMM',
    # Missing from API, is in website https://coinmarketcap.com/currencies/cryptcoin/
    'CRYPT',
    # Missing from API, is in https://coinmarketcap.com/currencies/cartesi/
    'CTSI',
    # Missing from API, is in https://coinmarketcap.com/currencies/conspiracycoin/
    'CYC',
    # Missing from API, is in https://coinmarketcap.com/currencies/diem/
    'DIEM',
    # Missing from API, is in https://coinmarketcap.com/currencies/darkcash/
    'DRKC',
    # Missing from API, is in https://coinmarketcap.com/currencies/dashcoin/
    'DSH',
    # Missing from API, is in https://coinmarketcap.com/currencies/earthcoin/
    'EAC',
    # Missing from API, is in https://coinmarketcap.com/currencies/execoin/
    'EXE',
    # Missing from API, is in https://coinmarketcap.com/currencies/fantomcoin/
    'FCN',
    # Missing from API, is in https://coinmarketcap.com/currencies/fibre/
    'FIBRE',
    # Missing from API, is in https://coinmarketcap.com/currencies/flappycoin/
    'FLAP',
    # Missing from API, is in https://coinmarketcap.com/currencies/fluttercoin/
    'FLT',
    # Missing from API, is in https://coinmarketcap.com/currencies/fractalcoin/
    'FRAC',
    # Missing from API, is in https://coinmarketcap.com/currencies/franko/
    'FRK',
    # Missing from API, is in https://coinmarketcap.com/currencies/gapcoin/
    'GAP',
    # Missing from API, is in https://coinmarketcap.com/currencies/gems/
    'GEMZ',
    # Missing from API, is in https://coinmarketcap.com/currencies/gameleaguecoin/
    'GML',
    # Missing from API, is in https://coinmarketcap.com/currencies/gpucoin/
    'GPUC',
    # Missing from API, is in https://coinmarketcap.com/currencies/guerillacoin/
    'GUE',
    # Missing from API, is in https://coinmarketcap.com/currencies/hive-dollar/
    'HBD',
    # Missing from API, is in https://coinmarketcap.com/currencies/hive-blockchain/
    'HIVE',
    # Missing from API, is in https://coinmarketcap.com/currencies/bigcoin/
    'HUGE',
    # Missing from API, is in https://coinmarketcap.com/currencies/heavycoin/
    'HVC',
    # Missing from API, is in https://coinmarketcap.com/currencies/next-horizon/
    'HZ',
    # Missing from API, is in https://coinmarketcap.com/currencies/klondikecoin/
    'KDC',
    # Missing from API, is in https://coinmarketcap.com/currencies/keycoin/
    'KEY-3',
    # Missing from API, is in https://coinmarketcap.com/currencies/leafcoin/
    'LEAF',
    # Missing from CMC, is in cryptocompare: https://www.cryptocompare.com/coins/loon
    'LOON',
    # Missing from API, is in https://coinmarketcap.com/currencies/ltbcoin/
    'LTBC',
    # Missing from API, is in https://coinmarketcap.com/currencies/litecoinx/
    'LTCX',
    # Missing from API, is in https://coinmarketcap.com/currencies/monetaverde/
    'MCN',
    # Missing from API, is in https://coinmarketcap.com/currencies/minerals/
    'MIN',
    # Missing from API, is in https://coinmarketcap.com/currencies/memorycoin/
    'MMC',
    # Missing from API, is in https://coinmarketcap.com/currencies/mmnxt/
    'MMNXT',
    # Missing from API, is in https://coinmarketcap.com/currencies/mmxiv/
    'MMXIV',
    # Missing from API, is in https://coinmarketcap.com/currencies/mazacoin/
    'MAZA',
    # Missing from API, is in https://coinmarketcap.com/currencies/nautiluscoin/
    'NAUT',
    # Missing from API, is in https://coinmarketcap.com/currencies/noblecoin/
    'NOBL',
    # Missing from API, is in https://coinmarketcap.com/currencies/noirshares/
    'NRS',
    # Missing from API, is in https://coinmarketcap.com/currencies/nxtinspect/
    'NXTI',
    # Missing from API is https://coinmarketcap.com/currencies/polybit/
    'POLY-2',
    # Missing from API is https://coinmarketcap.com/currencies/prospercoin/
    'PRC',
    # Missing from API is https://coinmarketcap.com/currencies/prcoin/
    'PRC-2',
    # Missing from API is https://coinmarketcap.com/currencies/qubitcoin/
    'Q2C',
    # Missing from API is https://coinmarketcap.com/currencies/qibuck/
    # and https://coinmarketcap.com/currencies/qibuck-asset/
    'QBK',
    # Missing from API is https://coinmarketcap.com/currencies/quazarcoin-old/
    # There is also a new one but we don't support the symbol yet
    # https://coinmarketcap.com/currencies/quasarcoin/ (QAC)
    'QCN',
    # Missing from API is https://coinmarketcap.com/currencies/qora/
    'QORA',
    # Missing from API is https://coinmarketcap.com/currencies/quatloo/
    'QTL',
    # Missing from API is https://coinmarketcap.com/currencies/riecoin/
    'RIC',
    # Missing from API is https://coinmarketcap.com/currencies/razor/
    'RZR',
    # Missing from API is https://coinmarketcap.com/currencies/shadowcash/
    'SDC',
    # Missing from API is https://coinmarketcap.com/currencies/silkcoin/
    'SILK',
    # Missing from API is https://coinmarketcap.com/currencies/spaincoin/
    'SPA',
    # Squallcoin. Completely missing ... but is in cryptocompare
    'SQL',
    # Missing from API is https://coinmarketcap.com/currencies/sonicscrewdriver/
    'SSD',
    # Missing from API is https://coinmarketcap.com/currencies/swarm/
    'SWARM',
    # Missing from API is https://coinmarketcap.com/currencies/sync/
    'SYNC',
    # Missing from API is https://coinmarketcap.com/currencies/torcoin-tor/
    'TOR',
    # Missing from API is https://coinmarketcap.com/currencies/trustplus/
    'TRUST',
    # Missing from API is https://coinmarketcap.com/currencies/unitus/
    'UIS',
    # Missing from API is https://coinmarketcap.com/currencies/umbrella-ltc/
    'ULTC',
    # Missing from API is https://coinmarketcap.com/currencies/supernet-unity/
    'UNITY',
    # Missing from API is https://coinmarketcap.com/currencies/uro/
    'URO',
    # Missing from API is https://coinmarketcap.com/currencies/usde/
    'USDE',
    # Missing from API is https://coinmarketcap.com/currencies/utilitycoin/
    'UTIL',
    # Missing from API is https://coinmarketcap.com/currencies/vootcoin/
    'VOOT',
    # InsanityCoin (WOLF). Completely missing ... but is in cryptocompare
    'WOLF',
    # Missing from API is https://coinmarketcap.com/currencies/sapience-aifx/
    'XAI',
    # Missing from API is https://coinmarketcap.com/currencies/crypti/
    'XCR',
    # Missing from API is https://coinmarketcap.com/currencies/dogeparty/
    'XDP',
    # Missing from API is https://coinmarketcap.com/currencies/libertycoin/
    'XLB',
    # Missing from API is https://coinmarketcap.com/currencies/pebblecoin/
    'XPB',
    # Missing from API is https://coinmarketcap.com/currencies/stabilityshares/
    'XSI',
    # Missing from API is https://coinmarketcap.com/currencies/vcash/
    'XVC',
    # Missing from API is https://coinmarketcap.com/currencies/yaccoin/
    'YACC',
    # GlobalCoin. Missing from API is https://coinmarketcap.com/currencies/globalcoin/
    'GLC-2',
    # Bytecent. Missing from API is https://coinmarketcap.com/currencies/bytecent/
    'BYC',
    # Bytecent. Missing from API is https://coinmarketcap.com/currencies/apx/
    'APX',
    # RCoin. Missing from API is https://coinmarketcap.com/currencies/rcoin/
    'RCN-2',
    # Blazecoin. Missing from API is https://coinmarketcap.com/currencies/blazecoin/
    'BLZ-2',
    # BTE. Missing from coinmarketcap but is in cryptocompare:
    # https://www.cryptocompare.com/coins/bte/overview
    'BTE',
    # Bitgem. Missing from API is https://coinmarketcap.com/currencies/bitgem/
    'BTG-2',
    # Harvest Masternode coin. Missing from API but is in
    # https://coinmarketcap.com/currencies/harvest-masternode-coin/
    'HC-2',
    # Triggers. Missing from API but is in
    # https://coinmarketcap.com/currencies/triggers/
    'TRIG',
    # 300 token. Missing from API but is in
    # https://coinmarketcap.com/currencies/300-token/
    '300',
    # Accelerator Network. Missing from API but is in
    # https://coinmarketcap.com/currencies/accelerator-network/
    'ACC-3',
    # Amis. Missing from API bus it in
    # https://coinmarketcap.com/currencies/amis/
    'AMIS',
    # ArcadeCity token. Missing from API bus it in
    # https://coinmarketcap.com/currencies/arcade-token/
    'ARC-2',
    # Astronaut token. Missing from API bus it in
    # https://coinmarketcap.com/currencies/astro/
    'ASTRO',
    # Artex coin. Missing from API but is in
    # https://coinmarketcap.com/currencies/artex-coin/
    'ATX-2',
    # Avalon coin. Missing from API but is in cryptocompare
    # https://www.cryptocompare.com/coins/ava/overview
    'AVA-2',
    # BitAsean. Missing from API but is in
    # https://coinmarketcap.com/currencies/bitasean/
    'BAS',
    # BelugaPay. Missing from API but is in
    # https://coinmarketcap.com/currencies/belugapay/
    'BBI',
    # Blockchain.capital. Missing from API but is in
    # https://coinmarketcap.com/currencies/bcap/
    'BCAP',
    # Bankcoin BCash (https://bankcoinbcash.com/)
    # is not in coinmarketcap but is in paprika
    'BCASH',
    # BHPCash is in coinmarketcap only as the new native token BHP
    # https://coinmarketcap.com/currencies/bhp-coin/
    'BHPC',
    # BITCAR (https://www.coingecko.com/en/coins/bitcar)
    # is not in coinmarketcap but is in cryptocompare
    'BITCAR',
    # BITPARK (https://coinmarketcap.com/currencies/bitpark-coin/)
    # is not in coinmarketcap but is in paprika
    'BITPARK',
    # BankCoin Cash (https://bankcoin-cash.com/)
    # is not in coinmarketcap but is in paprika
    'BKC',
    # Blockchain Index (https://coinmarketcap.com/currencies/blockchain-index/)
    # is not in coinmarketcap but is in cryptocompare
    'BLX',
    # BMChain (https://www.cryptocompare.com/coins/bmt/overview)
    # is not in coinmarketcap but is in cryptocompare
    'BMT',
    # Boulle (https://www.cryptocompare.com/coins/bou/overview)
    # is not in coinmarketcap but is in cryptocompare
    'BOU',
    # Bitair (https://coinmarketcap.com/currencies/bitair/)
    # is not in coinmarketcap but is in cryptocompare
    'BTCA',
    # EthereumBitcoin (https://www.cryptocompare.com/coins/btce/overview)
    # is not in coinmarketcap but is in cryptocompare
    'BTCE',
    # Bytether (https://www.cryptocompare.com/coins/byther/overview)
    # is not in coinmarketcap but is in cryptocompare
    'BTH',
    # Bither (https://www.cryptocompare.com/coins/btr/)
    # is not in coinmarketcap but is in cryptocompare
    'BTR-2',
    # Dice Money (https://www.cryptocompare.com/coins/cetstar/overview)
    # is not in coinmarketcap but is in cryptocompare
    'CET-2',
    # Cofoundit. Missing from API bus it in
    # https://coinmarketcap.com/currencies/cofound-it/
    'CFI',
    # Crafty (https://www.cryptocompare.com/coins/cfty/overview)
    # is not in coinmarketcap but is in cryptocompare
    'CFTY',
    # Climatecoin (https://www.cryptocompare.com/coins/co2/overview)
    # is not in coinmarketcap but is in cryptocompare
    'CO2',
    # CoinPulseToken (https://www.cryptocompare.com/coins/cpex/overview)
    # is not in coinmarketcap but is in cryptocompare
    'CPEX',
    # CrowdCoin (https://www.cryptocompare.com/coins/crcstarstarstar/overview)
    # is not in coinmarketcap but is in cryptocompare
    'CRC-2',
    # CargoCoin (https://www.cryptocompare.com/coins/crgo/overview)
    # is not in coinmarketcap but is in cryptocompare
    'CRGO',
    # Centra missing from API but is in
    # https://coinmarketcap.com/currencies/centra/
    'CTR',
    # CarTaxi (https://www.cryptocompare.com/coins/ctx/overview)
    # is not in coinmarketcap but is in cryptocompare
    'CTX-2',
    # Digital Developers Fund (https://coinmarketcap.com/currencies/digital-developers-fund/)
    # is not in coinmarketcap but is in cryptocompare
    'DDF',
    # Depository Network (https://www.cryptocompare.com/coins/depo/overview)
    # is not in coinmarketcap but is in cryptocompare
    'DEPO',
    # DigiPulse missing from API but is in
    # https://coinmarketcap.com/currencies/digipulse/
    'DGPT',
    # Etherisc (https://www.cryptocompare.com/coins/dip/overview)
    # is not in coinmarketcap but is in cryptocompare
    'DIP',
    # Divi Exchange Token (https://www.cryptocompare.com/coins/divx/)
    # is not in coinmarketcap but is in cryptocompare
    'DIVX',
    # Digital Assets Power Play (https://www.cryptocompare.com/coins/dpp/)
    # is not in coinmarketcap but is in cryptocompare
    'DPP',
    # DreamTeam Token (https://www.cryptocompare.com/coins/dream/)
    # is not in coinmarketcap but is in cryptocompare
    'DREAM',
    # Dcorp (https://www.cryptocompare.com/coins/drp/)
    # is not in coinmarketcap but is in cryptocompare
    'DRP',
    # DigitalTicks (https://www.coingecko.com/en/coins/digital-ticks)
    # is not in coinmarketcap but is in paprika
    'DTX-2',
    # Decentralized Universal Basic Income (https://www.cryptocompare.com/coins/dubi/)
    # is not in coinmarketcap but is in cryptocompare
    'DUBI',
    # E4ROW missing from API but is in
    # https://coinmarketcap.com/currencies/ether-for-the-rest-of-the-world/
    'E4ROW',
    # EAGLE missing from API but is in
    # https://coinmarketcap.com/currencies/eaglecoin/
    'EAGLE',
    # eBitcoinCash missing from API but is in
    # https://coinmarketcap.com/currencies/ebitcoin-cash/
    'EBCH',
    # OpenSource university (https://os.university/) is not in
    # coinmarketcap but is in paprika
    'EDU-2',
    # EGAS missing from API but is in
    # https://coinmarketcap.com/currencies/ethgas/
    'EGAS',
    # EasyMine (https://www.cryptocompare.com/coins/emt/overview)
    # is not in coinmarketcap but is in cryptocompare
    'EMT',
    # Ethereum Movie Venture missing from API but is in
    # https://coinmarketcap.com/currencies/ethereum-movie-venture/
    'EMV',
    # Hut34 Entropy token (https://www.cryptocompare.com/coins/entrp/overview)
    # is not in coinmarketcap but is in cryptocompare
    'ENTRP',
    # EtherBTC (https://www.cryptocompare.com/coins/ethb/overview)
    # is not in coinmarketcap but is in cryptocompare
    'ETHB',
    # Ethereum Dark missing from API but is in
    # https://coinmarketcap.com/currencies/ethereum-dark/
    'ETHD',
    # EximChain missing from API but is in
    # https://coinmarketcap.com/currencies/eximchain/
    'EXC-2',
    # Fingerprint  (https://fingerprintcoin.org/) is not
    # in coinmarketcap but is in paprika
    'FGP',
    # FidelityHouse (https://www.cryptocompare.com/coins/fih/overview)
    # is not in coinmarketcap but is in cryptocompare
    'FIH',
    # GastroAdvisor (https://www.cryptocompare.com/coins/fork/)
    # is not in coinmarketcap but is in cryptocompare. Note. Coinmarketcap
    # has the same symbol (FORK) for another asset:
    # https://coinmarketcap.com/currencies/forkcoin/#charts
    'FORK-2',
    # Farad missing from API but is in
    # https://coinmarketcap.com/currencies/farad/
    'FRD',
    # Fitrova missing from API but is in
    # https://coinmarketcap.com/currencies/fitrova/
    'FRV',
    # Finally Usable Crypto Karma token (FUCK) missing from API but is in
    # https://coinmarketcap.com/currencies/fucktoken/
    'FUCK',
    # Fund Yourself Now (FYN) missing from API but is in
    # https://coinmarketcap.com/currencies/fundyourselfnow/
    'FYN',
    # Globitex (https://www.cryptocompare.com/coins/gbxt/overview)
    # is not in coinmarketcap but is in cryptocompare
    'GBX-2',
    # Gimli missing from API but is in
    # https://coinmarketcap.com/currencies/gimli/
    'GIM',
    # Gladius token missing from API but is in
    # https://coinmarketcap.com/currencies/gladius-token/
    'GLA',
    # Mercury Protocol missing from API but is in
    # https://coinmarketcap.com/currencies/mercury-protocol/
    'GMT',
    # Hawala exchange token missing from API but is in
    # https://coinmarketcap.com/currencies/hat-exchange/
    'HAT',
    # Hedge token missing from API but is in
    # https://coinmarketcap.com/currencies/hedge/
    'HDG',
    # HackerGold token missing from API but is in
    # https://coinmarketcap.com/currencies/hacker-gold/
    'HKG',
    # iDice token missing from API but is in
    # https://coinmarketcap.com/currencies/idice/
    'ICE',
    # ICOS token missing from API but is in
    # https://coinmarketcap.com/currencies/icos/
    'ICOS',
    # Intimate. Completely missing ... but is in cryptocompare
    'ITM',
    # JOYSO (https://www.cryptocompare.com/coins/joystar/overview)
    # is not in coinmarketcap but is in cryptocompare
    'JOY',
    # Kuende (https://www.cryptocompare.com/coins/kue/overview)
    # is not in coinmarketcap but is in cryptocompare
    'KUE',
    # Lendconnect is missing from API but is in
    # https://coinmarketcap.com/currencies/lendconnect/
    'LCT',
    # Liquorchain (https://etherscan.io/address/0x4A37A91eec4C97F9090CE66d21D3B3Aadf1aE5aD)
    # is not in coinmarketcap but is in paprika
    'LCT-2',
    # Logarithm (https://www.cryptocompare.com/coins/lgr/overview)
    # is not in coinmarketcap but is in cryptocompare
    'LGR',
    # Link Platform missing from API but is in
    # https://coinmarketcap.com/currencies/link-platform/
    'LNK',
    # Live Start missing from API but is in
    # https://coinmarketcap.com/currencies/live-stars/
    'LIVE',
    # Linker Coin  missing from API but is in
    # https://coinmarketcap.com/currencies/linker-coin/
    'LNC-2',
    # Locus Chain (https://etherscan.io/address/0xC64500DD7B0f1794807e67802F8Abbf5F8Ffb054)
    # is not in coinmarketcap but is in paprika
    'LOCUS',
    # Embers (https://coinmarketcap.com/currencies/embers/)
    # is not in coinmarketcap but is in cryptocompare
    'MBRS',
    # Musiconomi (https://coinmarketcap.com/currencies/musiconomi/)
    # is not in coinmarketcap but is in cryptocompare
    'MCI',
    # All.me (https://www.cryptocompare.com/coins/me/overview)
    # is not in coinmarketcap but is in cryptocompare
    'ME',
    # Meshbox (https://coinlib.io/coin/MESH/MeshBox)
    # is not in coinmarketcap but is in paprika
    'MESH',
    # Micro Licensing Coin (https://www.cryptocompare.com/coins/milc/overview)
    # is not in coinmarketcap but is in cryptocompare
    'MILC',
    # Media Network Token (https://www.cryptocompare.com/coins/mntstar/overview)
    # is not in coinmarketcap but is in cryptocompare
    'MNT',
    # Money Rebel (https://www.cryptocompare.com/coins/mrpstar/overview)
    # is not in coinmarketcap but is in cryptocompare
    'MRP',
    # Macroverse (https://www.cryptocompare.com/coins/mrv/overview)
    # is not in coinmarketcap but is in cryptocompare
    'MRV',
    # Mothership missing from API but is in
    # https://coinmarketcap.com/currencies/mothership/
    'MSP',
    # Nami ICO (https://etherscan.io/address/0x8d80de8A78198396329dfA769aD54d24bF90E7aa)
    # is not in coinmarketcap but is in paprika
    'NAC',
    # NeedsCoin (https://etherscan.io/address/0x9344b383b1d59b5ce3468b234dab43c7190ba735)
    # is not in coimarketcap but is in paprika
    'NCC-2',
    # Newbium missing from API but is in
    # https://coinmarketcap.com/currencies/newbium/
    'NEWB',
    # Nimfamoney missing from API but is in
    # https://coinmarketcap.com/currencies/nimfamoney/
    'NIMFA',
    # Network Token missing from API but is in
    # https://coinmarketcap.com/currencies/network-token/
    'NTWK',
    # Nexxus missing from API but is in
    # https://coinmarketcap.com/currencies/nexxus/
    'NXX',
    # Acorn Collective (https://www.cryptocompare.com/coins/oak/overview)
    # is not in coinmarketcap but is in cryptocompare
    'OAK',
    # Original Crypto coin (https://www.cryptocompare.com/coins/occ/overview)
    # is not in coinmarketcap but is in cryptocompare
    'OCC-2',
    # Wisepass missing from API but is in
    # https://coinmarketcap.com/currencies/wisepass/
    'PASS-2',
    # Publica missing from API but is in
    # https://coinmarketcap.com/currencies/publica/
    'PBL',
    # POP Chest Token (https://www.cryptocompare.com/coins/popc/overview)
    # is not in coinmarketcap but is in cryptocompare
    'POP-2',
    # Oyster missing from API but is in
    # https://coinmarketcap.com/currencies/oyster/
    'PRL',
    # Purpose missing from API but is in
    # https://coinmarketcap.com/currencies/purpose/
    'PRPS',
    # Quantum missing from API but is in
    # https://coinmarketcap.com/currencies/quantum/
    'QAU',
    # Qvolta missing from API but is in
    # https://coinmarketcap.com/currencies/qvolta/
    'QVT',
    # Realisto (https://www.cryptocompare.com/coins/rea/overview)
    # is not in coinmarketcap but is in cryptocompare
    'REA',
    # Red Cab (https://www.cryptocompare.com/coins/redc/overview)
    # is not in coinmarketcap but is in cryptocompare
    'REDC',
    # Rusgas missing from API but is in
    # https://coinmarketcap.com/currencies/rusgas/
    'RGS',
    # Riptide Coin (https://www.cryptocompare.com/coins/ript/overview)
    # is not in coinmarketcap but is in cryptocompare
    'RIPT',
    # RemiCoin missing from API but is in
    # https://coinmarketcap.com/currencies/remicoin/
    'RMC',
    # Render Token (https://www.cryptocompare.com/coins/rndr/overview)
    # is not in coinmarketcap but is in cryptocompare
    'RNDR',
    # RasputinOnlineCoin missing from API but is in
    # https://coinmarketcap.com/currencies/rasputin-online-coin/
    'ROC',
    # SGPay (https://www.cryptocompare.com/coins/sgp/overview)
    # is not in coinmarketcap but is in cryptocompare
    'SGP',
    # OysterShell missing from API but is in
    # https://coinmarketcap.com/currencies/oyster-shell/
    'SHL',
    # Smart Investment Fund Token missing from API but is in
    # https://coinmarketcap.com/currencies/smart-investment-fund-token/
    'SIFT',
    # SkrillaToken (https://www.cryptocompare.com/coins/skrstar/overview)
    # is not in coinmarketcap but is in cryptocompare
    'SKR',
    # Skraps (https://www.cryptocompare.com/coins/skrp/overview)
    # is not in coinmarketcap but is in cryptocompare
    'SKRP',
    # SkyMap (https://www.cryptocompare.com/coins/skym/overview)
    # is not in coinmarketcap but is in cryptocompare
    'SKYM',
    # SmartBillion (https://www.cryptocompare.com/coins/smartstar/overview)
    # is not in coinmarketcap but is in cryptocompare
    'SMART-2',
    # snowball (https://etherscan.io/address/0x198A87b3114143913d4229Fb0f6D4BCb44aa8AFF)
    # is not in coinmarketcap but is in paprika
    'SNBL',
    # SandCoin (https://www.cryptocompare.com/coins/snd/overview)
    # is not in coinmarketcap but is in paprika. SND in coinmarketcap is SnodeCoin
    'SND',
    # Spice VC Token (https://www.cryptocompare.com/coins/spice/overview)
    # is not in coinmarketcap but is in paprika
    'SPICE',
    # StreamSpace (https://www.cryptocompare.com/coins/ssh/overview)
    # is not in coinmarketcap but is in cryptocompare
    'SSH',
    # StashPay (https://www.cryptocompare.com/coins/stp/overview)
    # is not in coinmarketcap but is in cryptocompare
    'STP',
    # StarCredits missing from API but is in
    # https://coinmarketcap.com/currencies/starcredits/
    'STRC',
    # Taklimakan (https://www.cryptocompare.com/coins/tan/overview)
    # is not in coinmarketcap but is in cryptocompare
    'TAN',
    # T-Bot (https://www.cryptocompare.com/coins/tbt/overview)
    # is not in coinmarketcap but is in cryptocompare
    'TBT',
    # TercetNetwork (https://etherscan.io/address/0x28d7F432d24ba6020d1cbD4f28BEDc5a82F24320)
    # is not in coinmarketcap but is in paprika
    'TCNX',
    # Trade.io (https://www.cryptocompare.com/coins/tio/)
    # is not in coinmarketcap but is in paprika and cryptocompare
    'TIO',
    # CryptoInsight is missing from API but is in
    # https://coinmarketcap.com/currencies/trackr/
    'TKR',
    # Missing from API but are both in:
    # https://coinmarketcap.com/currencies/3x-short-trx-token/
    # https://coinmarketcap.com/currencies/3x-long-trx-token/
    'TRXBEAR',
    'TRXBULL',
    # TaTaTu is missing from API but is in
    # https://coinmarketcap.com/currencies/tatatu/
    'TTU',
    # Urbit (https://www.cryptocompare.com/coins/urb/overview)
    # is not in coinmarketcap but is in cryptocompare
    'URB',
    'UTI',
    # USDJ is missing from coinmarketcap but is in cryptocompare
    # https://www.cryptocompare.com/coins/usdj/overview
    'USDJ',
    # Bitcoin Card (https://etherscan.io/address/0x9a9bB9b4b11BF8eccff84B58a6CCCCD4058A7f0D)
    # is not in coinmarketcap but is in paprika
    'VD',
    # VenusEnergy (https://www.cryptocompare.com/coins/venus/overview)
    # is not in coinmarketcap but is in cryptocompare
    'VENUS',
    # WhenCoin (https://www.cryptocompare.com/coins/when/overview)
    # is not in coinmarketcap but is in cryptocompare
    'WHEN',
    # WildCrypto missing from API but is in
    # https://coinmarketcap.com/currencies/wild-crypto/
    'WILD',
    # Wcoin missing from API but is in
    # https://coinmarketcap.com/currencies/wawllet/
    'WIN-2',
    # WeMark (https://www.cryptocompare.com/coins/wmk/overview)
    # is not in coinmarketcap but is in cryptocompare
    'WMK',
    # Wolk (https://www.cryptocompare.com/coins/wlk/overview)
    # is not in coinmarketcap but is in cryptocompare
    'WLK',
    # Gigawatt missing from API but is in
    # https://coinmarketcap.com/currencies/giga-watt-token/
    'WTT',
    # Sphere Identity missing from API but is in
    # https://coinmarketcap.com/currencies/sphere-identity/
    'XID',
    # Rialto missing from API but is in
    # https://coinmarketcap.com/currencies/rialto/
    'XRL',
    # ZIX (https://www.cryptocompare.com/coins/zix/)
    # is not in coinmarketcap but is in cryptocompare
    'ZIX',
)


# There can be multiple ids for the same symbol and for cases such as this
# we use this mapping to manually map Rotkehlchen symbols to CMC IDs
WORLD_TO_CMC_ID = {
    # For Rotkehlchen AID is AidCoin and AID-2 is Aidus token
    'AID': 2462,
    'AID-2': 3785,
    'ALQO': 2199,
    # For Rotkehlchen ATOM is Cosmos and ATOM-2 is Atomic Coin
    'ATOM': 3794,
    'ATOM-2': 1420,
    # Bitstars
    'BITS': 276,
    # Bitswift
    'BITS-2': 659,
    # Bitmark
    'BTM': 543,
    # Bytom
    'BTM-2': 1866,
    # 3x Long Bitcoin
    'BULL': 5157,
    # DEC is Darico Ecosystem Coin and DEC-2 is Decenter
    'DEC': 3376,
    'DEC-2': 5835,
    # FairCoin
    'FAIR': 224,
    # FairGame
    'FAIR-2': 2366,
    # Selfkey
    'KEY': 2398,
    # KEY (bihu)
    'KEY-2': 2713,
    # KyberNetwork
    'KNC': 1982,
    # KingN Coin
    'KNC-2': 1743,
    # Goldcoind (GLC in Rotkehlcen) is in coinmarketcap with ID 25
    # and symbol GLD. Symbol discrepancy is the same as in cryptocompare
    'GLC': 25,
    # For Rotkehlchen MORE is More coin and MORE-2 is Mithril Ore token
    'MORE': 1722,
    'MORE-2': 3206,
    # For Rotkehlchen CBC is Cashbet Coin and CBC-2 is Cashbery coin
    'CBC': 2855,
    'CBC-2': 3199,
    # For Rotkehlcen CMCT is Crown machine token and CMCT-2 is Cyber movie chain token
    'CMCT': 2708,
    'CMCT-2': 3429,
    # For Rotkehlchen EDR is Endor Protocol and EDR-2 is E-Dinar coin
    'EDR': 2835,
    'EDR-2': 1358,
    # For Rotkehlchen USDS is Stable USD and USDS-2 is stronghold usd
    'USDS': 3719,
    'USDS-2': 3641,
    # For Rotkehlchen BTT is BitTorrent and BTT-2 is blocktrade token
    'BTT': 3718,
    'BTT-2': 3084,
    # For Rotkehlchen ONG is Ontology gas and ONG-2 is SoMee.Social
    'ONG': 3217,
    'ONG-2': 2240,
    # For Rotkehlchen SLT is SmartLands and SLT-2 is Social Lending Network
    'SLT': 2471,
    'SLT-2': 3117,
    # For Rotkehlchen PAI is Project Pai and PAI-2 is PCHAIN
    'PAI': 2900,
    'PAI-2': 2838,
    # For Rotkehlchen CMT is CyberMiles and CMT-2 is CometCoin
    'CMT': 2246,
    'CMT-2': 1291,
    # For Rotkehlchen HOT is Holochain and HOT-2 is Hydro Protocol
    'HOT': 2682,
    'HOT-2': 2430,
    # For Rotkehlchen IOTA is IOTA but in coimarketcap it's MIOTA
    'IOTA': 1720,
    # For Rotkehlchen YOYOW if YOYO but in coinmarketcap it's YOYOW
    'YOYOW': 1899,
    # For Rotkehlchen ACC is AdCoin, ACC-2 is ACChain and ACC-3 is Accelerator network
    'ACC': 1915,
    'ACC-2': 2515,
    # For Rotkehlchen ADST is Adshares but in coinmarketcap it's ADS
    'ADST': 1883,
    # For Rotkehlchen ARB is Arbitrage coin and ARB-2 is ARbit
    'ARB': 2985,
    'ARB-2': 938,
    # Symbol for B2BX is B2B is coinmarketcap so we need to specify it
    'B2BX': 2204,
    # For Rotkehlchen BBK is BrickBlock and BBK-2 is Bitblocks
    'BBK': 3015,
    'BBK-2': 3051,
    # For Rotkehlchen BET is Dao.Casino and BET-2 is BetaCoin
    'BET': 1771,
    'BET-2': 70,
    # For Rotkehlchen BLOC is BlockCloud and BLOC-2 is Bloc.Money
    'BLOC': 3860,
    'BLOC-2': 3451,
    # For Rotkehlchen BOX is BoxToken and BOX-2 is ContentBox
    'BOX': 3475,
    'BOX-2': 2945,
    # For Rotkehlchen CAN is CanYaCoin and CAN-2 is Content And Ad Network
    'CAN': 2343,
    'CAN-2': 2358,
    # For Rotkehlchen CAT is Bitclave and CAT-2 is BlockCAT
    'CAT': 2334,
    'CAT-2': 1882,
    # For Rotkehlchen COS is Contentos. If ever want to add the other COS it's 1989
    'COS': 4036,
    # For Rotkehlchen CPC is CPChain and CPC-2 is CapriCoin
    'CPC': 2482,
    'CPC-2': 1008,
    # For Rotkehlchen EVN is Envion and EVN-2 is EvenCoin
    'EVN': 2526,
    'EVN-2': 3261,
    # For Rotkehlchen FT is Fabric Token and FT-2 is FCoin
    'FT': 2768,
    'FT-2': 2904,
    # For Rotkehlchen GENE is ParkGene and GENE-2 is Gene Source Code Chain
    'GENE': 2720,
    'GENE-2': 3297,
    # For Rotkehlchen GET is Guaranteed Entrance Token and GET-2 is Themis
    'GET': 2354,
    'GET-2': 3127,
    # For Rotkehlchen GOT is Go Network Token and GOT-2 is ParkinGo
    'GOT': 2898,
    'GOT-2': 3251,
    # For Rotkehlchen HMC is Hi Mutual Society and HMC-2 is Harmony Coin
    'HMC': 2484,
    'HMC-2': 1832,
    # For Rotkehlchen HMC is Hi Mutual Society and HMC-2 is Harmony Coin
    # For Rotkehlchen IQ is Everipedia and IQ-2 is IQ.cash
    'IQ': 2930,
    'IQ-2': 3273,
    'KNT': 3086,
    'KNT-2': 3383,
    # For Rotkehlchen LUNA is Luna Coin and LUNA-2 is Terra Luna
    'LUNA': 1496,
    'LUNA-2': 4172,
    # For Rotkehlchen MTC is doc.com Token and MTC-2 is Mesh Network
    'MTC': 2711,
    'MTC-2': 2936,
    # For Rotkehlchen NTK is Neurotoken and NTK-2 is NetKoin
    'NTK': 2536,
    'NTK-2': 3149,
    # For Rotkehlchen ONE is Menlo One and ONE-2 is Harmony
    'ONE': 3603,
    'ONE-2': 3945,
    # For Rotkehlchen ORS is OriginSport Token and ORS-2 is ORS Group
    'ORS': 2879,
    'ORS-2': 2911,
    # For Rotkehlchen PLA is Plair and PLA-2 is Playchip
    'PLA': 3711,
    'PLA-2': 3731,
    # For Rotkehlchen PNT is pNetwork and PNT-2 Penta
    'PNT': 5794,
    'PNT-2': 2691,
    # For Rotkehlcen SOL is Sola token and SOL-2 is Solana
    'SOL': 3333,
    'SOL-2': 5426,
    # For Rotkehlchen SOUL is Phantasma and SOUL-2 is CryptoSoul
    'SOUL': 2827,
    'SOUL-2': 3501,
    # For Rotkehlchen SPD is Spindle and SPD-2 is Stipend
    'SPD': 2828,
    'SPD-2': 2616,
    # For Rotkehlchen TCH is ThoreCash and TCH-2 is TigerCash
    'TCH': 3056,
    'TCH-2': 3806,
    # For Rotkehlchen TNC is TNC Token and we don't deal with Trinity network yet (2443)
    'TNC': 5524,
    # For Rotkehlchen WEB is Webcoin and WEB-2 Webchain
    'WEB': 3027,
    'WEB-2': 3361,
}


def find_cmc_coin_data(
        asset_symbol: str,
        cmc_list: List[Dict[str, Any]],
) -> Optional[Dict[str, Any]]:
    """Given an asset's symbol find its data in the coinmarketcap list"""
    if not cmc_list:
        return None

    if asset_symbol in WORLD_TO_CMC_ID:
        coin_id = WORLD_TO_CMC_ID[asset_symbol]
        for coin in cmc_list:
            if coin['id'] == coin_id:
                return coin

        raise AssertionError('The CMC id should always be correct. Is our data corrupt?')

    if asset_symbol in KNOWN_TO_MISS_FROM_CMC:
        return None

    found_coin_data: Optional[Dict[str, Any]] = None
    for coin in cmc_list:
        if coin['symbol'] == asset_symbol:
            if found_coin_data:
                print(
                    f'Asset with symbol {asset_symbol} was found in coinmarketcap '
                    f'both as {found_coin_data["id"]} - {found_coin_data["name"]} '
                    f'and {coin["id"]} - {coin["name"]}',
                )
                sys.exit(1)
            found_coin_data = coin

    if not found_coin_data:
        print(
            f"Could not find asset with canonical symbol {asset_symbol} in "
            f"coinmarketcap's coin list",
        )
        sys.exit(1)

    return found_coin_data


class Coinmarketcap():

    def __init__(self, data_directory: Path, api_key: str):
        self.prefix = 'https://pro-api.coinmarketcap.com/'
        self.backoff_limit = 180
        self.data_directory = data_directory
        self.session = requests.session()
        # As per coinmarketcap's API
        self.session.headers.update({
            'User-Agent': 'rotkehlchen',
            'X-CMC_PRO_API_KEY': api_key,
            'Accept': 'application/json',
            'Accept-Encoding': 'deflate, gzip',
        })

    def _query(self, path: str) -> str:
        backoff = INITIAL_BACKOFF
        while True:
            response = self.session.get(f'{self.prefix}{path}')
            if response.status_code == 429 and backoff < self.backoff_limit:
                gevent.sleep(backoff)
                backoff *= 2
                continue
            elif response.status_code != 200:
                raise RemoteError(
                    f'Coinpaprika API request {response.url} for {path} failed '
                    f'with HTTP status code {response.status_code} and text '
                    f'{response.text}',
                )

            return response.text

    def _get_cryptocyrrency_map(self) -> List[Dict[str, Any]]:
        start = 1
        limit = 5000
        result: List[Dict[str, Any]] = []
        while True:
            response_data = rlk_jsonloads_dict(
                self._query(f'v1/cryptocurrency/map?start={start}&limit={limit}'),
            )
            result.extend(response_data['data'])
            if len(response_data['data']) != limit:
                break

        return result

    def get_cryptocyrrency_map(self) -> List[Dict[str, Any]]:
        # TODO: Both here and in cryptocompare the cache funcionality is the same
        # Extract the caching part into its own function somehow and abstract it
        # away
        invalidate_cache = True
        coinlist_cache_path = os.path.join(self.data_directory, 'cmc_coinlist.json')
        if os.path.isfile(coinlist_cache_path):
            log.info('Found coinmarketcap coinlist cache', path=coinlist_cache_path)
            with open(coinlist_cache_path, 'r') as f:
                try:
                    file_data = rlk_jsonloads_dict(f.read())
                    now = ts_now()
                    invalidate_cache = False

                    # If we got a cache and it's over a month old then requery coinmarketcap
                    if file_data['time'] < now and now - file_data['time'] > 2629800:
                        log.info('Coinmarketcap coinlist cache is now invalidated')
                        invalidate_cache = True
                except JSONDecodeError:
                    invalidate_cache = True

        if invalidate_cache:
            data = self._get_cryptocyrrency_map()
            # Also save the cache
            with open(coinlist_cache_path, 'w') as f:
                now = ts_now()
                log.info('Writing coinmarketcap coinlist cache', timestamp=now)
                write_data = {'time': now, 'data': data}
                f.write(rlk_jsondumps(write_data))
        else:
            # in any case take the data
            data = file_data['data']

        return data



# -*- coding: utf-8 -*-
"""
Created on Fri May 31 08:17:21 2019

@author: Martin
"""

import configparser
import re
from pathlib import Path
import koalafolio.gui.QLogger as logger
import koalafolio.gui.QSettings as settings
import os

localLogger = logger.globalLogger

class ExportTranslator(configparser.ConfigParser):
    def __init__(self, dataPath="", *args, **kwargs):
        super(ExportTranslator, self).__init__(self, *args, **kwargs)

        if dataPath:
            self.setPath(dataPath)

    def setPath(self, path):
        self.filePath = os.path.join(path, 'translation.txt')
        # init settings
        self.initTranslation()
        if not os.path.isfile(self.filePath):
            self.saveTranslation()
        else:
            self.readTranslation()
            self.saveTranslation()
        return self

    def initTranslation(self):
        # set default settings
        # en
        self['en'] = {}
        self['en']['Profit'] = 'Profit'
        self['en']['Timeframe'] = 'Timeframe'
        self['en']['in'] = 'in'
        self['en']['Group'] = 'Group'
        self['en']['Buy'] = 'Buy'
        self['en']['Sell'] = 'Sell'
        self['en']['Date'] = 'Date'
        self['en']['Amount'] = 'Amount'
        self['en']['Price'] = 'Price'
        self['en']['Value'] = 'Value'
        self['en']['tax_relevant'] = 'tax relevant'
        self['en']['Fees'] = 'Fees'
        self['en']['Fee'] = 'Fee'
        self['en']['pc'] = 'pc'
        # de
        self['de'] = {}
        self['de']['Profit'] = 'Gewinn'
        self['de']['Timeframe'] = 'Zeitraum'
        self['de']['in'] = 'in'
        self['de']['Group'] = 'Gruppe'
        self['de']['Buy'] = 'Ankauf'
        self['de']['Sell'] = 'Verkauf'
        self['de']['Date'] = 'Datum'
        self['de']['Amount'] = 'Anzahl'
        self['de']['Price'] = 'Preis'
        self['de']['Value'] = 'Wert'
        self['de']['tax_relevant'] = 'zu versteuern'
        self['de']['Fees'] = 'Gebühren'
        self['de']['Fee'] = 'Gebühr'
        self['de']['pc'] = 'stk'

    def saveTranslation(self):
        try:
            with open(self.filePath, 'w') as configfile:
                self.write(configfile)
            logger.globalLogger.info('translation saved')
            return True
        except Exception as ex:
            logger.globalLogger.error('error in saveTranslation: ' + str(ex))
            return False

    def readTranslation(self):
        try:
            self.read(self.filePath)
            logger.globalLogger.info('translation loaded')
        except Exception as ex:
            logger.globalLogger.error('translations can not be loaded: ' + str(ex))
            self.resetDefault()

    def resetDefault(self):
        try:
            os.remove(self.filePath)
            self.initTranslation()
            self.saveTranslation()
            logger.globalLogger.info('translations reset to default')
        except Exception as ex:
            localLogger.warning('error resetting translations: ' + str(ex))

    def getLanguages(self):
        return [l for l in self if l != "DEFAULT"]

    def translate(self, text, lang):
        try:
            return self[lang][text.replace(' ', '_')]
        except:
            return text


from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from rotkehlchen.db.dbhandler import DBHandler


def _delete_bittrex_data(db: 'DBHandler') -> None:
    cursor = db.conn.cursor()
    # This is the user credentials trades table at v10
    cursor.execute('DELETE FROM trades WHERE location="D";')
    cursor.execute('DELETE FROM used_query_ranges WHERE name LIKE "bittrex_%";')
    db.conn.commit()


def upgrade_v11_to_v12(db: 'DBHandler') -> None:
    """Upgrades the DB from v11 to v12

    - Deletes all bittrex related DB data
    """
    _delete_bittrex_data(db)



import glob
import logging
import os
import re
from json.decoder import JSONDecodeError
from pathlib import Path
from typing import Any, Dict, Iterable, Iterator, List, NamedTuple, NewType, Optional

import gevent
import requests
from typing_extensions import Literal

from rotkehlchen.assets.asset import Asset, EthereumToken
from rotkehlchen.constants import ZERO
from rotkehlchen.db.dbhandler import DBHandler
from rotkehlchen.errors import (
    NoPriceForGivenTimestamp,
    PriceQueryUnsupportedAsset,
    RemoteError,
    UnsupportedAsset,
)
from rotkehlchen.externalapis.interface import ExternalServiceWithApiKey
from rotkehlchen.fval import FVal
from rotkehlchen.history import PriceHistorian
from rotkehlchen.logging import RotkehlchenLogsAdapter
from rotkehlchen.typing import ExternalService, Price, Timestamp
from rotkehlchen.utils.misc import (
    convert_to_int,
    timestamp_to_date,
    ts_now,
    write_history_data_in_file,
)
from rotkehlchen.utils.serialization import rlk_jsondumps, rlk_jsonloads_dict

logger = logging.getLogger(__name__)
log = RotkehlchenLogsAdapter(logger)


T_PairCacheKey = str
PairCacheKey = NewType('PairCacheKey', T_PairCacheKey)

RATE_LIMIT_MSG = 'You are over your rate limit please upgrade your account!'
CRYPTOCOMPARE_QUERY_RETRY_TIMES = 10
CRYPTOCOMPARE_SPECIAL_CASES_MAPPING = {
    Asset('TLN'): EthereumToken('WETH'),
    Asset('BLY'): EthereumToken('USDT'),
    Asset('cDAI'): EthereumToken('DAI'),
    Asset('cCOMP'): EthereumToken('COMP'),
    Asset('cBAT'): Asset('BAT'),
    Asset('cREP'): Asset('REP'),
    Asset('cSAI'): Asset('SAI'),
    Asset('cUSDC'): Asset('USDC'),
    Asset('cUSDT'): EthereumToken('USDT'),
    Asset('cWBTC'): Asset('WBTC'),
    Asset('cUNI'): Asset('UNI'),
    Asset('cZRX'): Asset('ZRX'),
    Asset('ADADOWN'): EthereumToken('USDT'),
    Asset('ADAUP'): EthereumToken('USDT'),
    Asset('BNBDOWN'): EthereumToken('USDT'),
    Asset('BNBUP'): EthereumToken('USDT'),
    Asset('BTCDOWN'): EthereumToken('USDT'),
    Asset('BTCUP'): EthereumToken('USDT'),
    Asset('ETHDOWN'): EthereumToken('USDT'),
    Asset('ETHUP'): EthereumToken('USDT'),
    Asset('EOSDOWN'): EthereumToken('USDT'),
    Asset('EOSUP'): EthereumToken('USDT'),
    Asset('DOTDOWN'): EthereumToken('USDT'),
    Asset('DOTUP'): EthereumToken('USDT'),
    Asset('LTCDOWN'): EthereumToken('USDT'),
    Asset('LTCUP'): EthereumToken('USDT'),
    Asset('TRXDOWN'): EthereumToken('USDT'),
    Asset('TRXUP'): EthereumToken('USDT'),
    Asset('XRPDOWN'): EthereumToken('USDT'),
    Asset('XRPUP'): EthereumToken('USDT'),
    Asset('DEXT'): EthereumToken('USDT'),
    Asset('DOS'): EthereumToken('USDT'),
    Asset('GEEQ'): EthereumToken('USDT'),
    Asset('LINKDOWN'): EthereumToken('USDT'),
    Asset('LINKUP'): EthereumToken('USDT'),
    Asset('XTZDOWN'): EthereumToken('USDT'),
    Asset('XTZUP'): EthereumToken('USDT'),
    Asset('STAKE'): EthereumToken('USDT'),
    Asset('MCB'): EthereumToken('USDT'),
    Asset('TRB'): EthereumToken('USDT'),
    Asset('YFI'): EthereumToken('USDT'),
    Asset('YAM'): EthereumToken('USDT'),
    Asset('DEC-2'): EthereumToken('USDT'),
    Asset('ORN'): EthereumToken('USDT'),
    Asset('PERX'): EthereumToken('USDT'),
    Asset('PRQ'): EthereumToken('USDT'),
    Asset('RING'): EthereumToken('USDT'),
    Asset('SBREE'): EthereumToken('USDT'),
    Asset('YFII'): EthereumToken('USDT'),
    Asset('BZRX'): EthereumToken('USDT'),
    Asset('CREAM'): EthereumToken('USDT'),
    Asset('ADEL'): EthereumToken('USDT'),
    Asset('ANK'): EthereumToken('USDT'),
    Asset('CORN'): EthereumToken('USDT'),
    Asset('SAL'): EthereumToken('USDT'),
    Asset('CRT'): EthereumToken('USDT'),
    Asset('FSW'): EthereumToken('USDT'),
    Asset('JFI'): EthereumToken('USDT'),
    Asset('PEARL'): EthereumToken('USDT'),
    Asset('TAI'): EthereumToken('USDT'),
    Asset('YFL'): EthereumToken('USDT'),
    Asset('TRUMPWIN'): EthereumToken('USDT'),
    Asset('TRUMPLOSE'): EthereumToken('USDT'),
    Asset('KLV'): EthereumToken('USDT'),
    Asset('KRT'): Asset('KRW'),
    Asset('RVC'): EthereumToken('USDT'),
    Asset('SDT'): EthereumToken('USDT'),
    Asset('CHI'): EthereumToken('USDT'),
    Asset('BAKE'): Asset('BNB'),
    Asset('BURGER'): Asset('BNB'),
    Asset('CAKE'): Asset('BNB'),
    Asset('BREE'): EthereumToken('USDT'),
    Asset('GHST'): EthereumToken('USDT'),
    Asset('MEXP'): EthereumToken('USDT'),
    Asset('POLS'): EthereumToken('USDT'),
    Asset('RARI'): EthereumToken('USDT'),
    Asset('VALUE'): EthereumToken('USDT'),
    Asset('$BASED'): EthereumToken('WETH'),
    Asset('DPI'): EthereumToken('WETH'),
    Asset('JRT'): EthereumToken('USDT'),
    Asset('PICKLE'): EthereumToken('USDT'),
    Asset('FILDOWN'): EthereumToken('USDT'),
    Asset('FILUP'): EthereumToken('USDT'),
    Asset('YFIDOWN'): EthereumToken('USDT'),
    Asset('YFIUP'): EthereumToken('USDT'),
    Asset('BOT'): EthereumToken('USDT'),
}
CRYPTOCOMPARE_SPECIAL_CASES = CRYPTOCOMPARE_SPECIAL_CASES_MAPPING.keys()


class PriceHistoryEntry(NamedTuple):
    time: Timestamp
    low: Price
    high: Price


class PriceHistoryData(NamedTuple):
    data: List[PriceHistoryEntry]
    start_time: Timestamp
    end_time: Timestamp


def _dict_history_to_entries(data: List[Dict[str, Any]]) -> List[PriceHistoryEntry]:
    """Turns a list of dict of history entries to a list of proper objects"""
    return [
        PriceHistoryEntry(
            time=Timestamp(entry['time']),
            low=Price(FVal(entry['low'])),
            high=Price(FVal(entry['high'])),
        ) for entry in data
    ]


def _dict_history_to_data(data: Dict[str, Any]) -> PriceHistoryData:
    """Turns a price history data dict entry into a proper object"""
    return PriceHistoryData(
        data=_dict_history_to_entries(data['data']),
        start_time=Timestamp(data['start_time']),
        end_time=Timestamp(data['end_time']),
    )


def _multiply_str_nums(a: str, b: str) -> str:
    """Multiples two string numbers and returns the result as a string"""
    return str(FVal(a) * FVal(b))


def pairwise(iterable: Iterable[Any]) -> Iterator:
    "s -> (s0, s1), (s2, s3), (s4, s5), ..."
    a = iter(iterable)
    return zip(a, a)


def _check_hourly_data_sanity(
        data: List[Dict[str, Any]],
        from_asset: Asset,
        to_asset: Asset,
) -> None:
    """Check that the hourly data is an array of objects having timestamps
    increasing by 1 hour.

    If not then a RemoteError is raised
    """
    index = 0
    for n1, n2 in pairwise(data):
        diff = n2['time'] - n1['time']
        if diff != 3600:
            raise RemoteError(
                'Unexpected fata format in cryptocompare query_endpoint_histohour. '
                "Problem at indices {} and {} of {}_to_{} prices. Time difference is: {}".format(
                    index, index + 1, from_asset, to_asset, diff),
            )

        index += 2


class Cryptocompare(ExternalServiceWithApiKey):
    def __init__(self, data_directory: Path, database: Optional[DBHandler]) -> None:
        super().__init__(database=database, service_name=ExternalService.CRYPTOCOMPARE)
        self.data_directory = data_directory
        self.price_history: Dict[PairCacheKey, PriceHistoryData] = {}
        self.price_history_file: Dict[PairCacheKey, Path] = {}
        self.session = requests.session()
        self.session.headers.update({'User-Agent': 'rotkehlchen'})

        # Check the data folder and remember the filenames of any cached history
        prefix = os.path.join(str(self.data_directory), 'price_history_')
        prefix = prefix.replace('\\', '\\\\')
        regex = re.compile(prefix + r'(.*)\.json')
        files_list = glob.glob(prefix + '*.json')

        for file_ in files_list:
            file_ = file_.replace('\\\\', '\\')
            match = regex.match(file_)
            assert match
            cache_key = PairCacheKey(match.group(1))
            self.price_history_file[cache_key] = Path(file_)

    def set_database(self, database: DBHandler) -> None:
        """If the cryptocompare instance was initialized without a DB this sets its DB"""
        msg = 'set_database was called on a cryptocompare instance that already has a DB'
        assert self.db is None, msg
        self.db = database

    def unset_database(self) -> None:
        """Remove the database connection from this cryptocompare instance

        This should happen when a user logs out"""
        msg = 'unset_database was called on a cryptocompare instance that has no DB'
        assert self.db is not None, msg
        self.db = None

    def _api_query(self, path: str) -> Dict[str, Any]:
        """Queries cryptocompare

        - May raise RemoteError if there is a problem reaching the cryptocompare server
        or with reading the response returned by the server
        """
        querystr = f'https://min-api.cryptocompare.com/data/{path}'
        log.debug('Querying cryptocompare', url=querystr)
        api_key = self._get_api_key()
        if api_key:
            querystr += f'&api_key={api_key}'

        tries = CRYPTOCOMPARE_QUERY_RETRY_TIMES
        while tries >= 0:
            try:
                response = self.session.get(querystr)
            except requests.exceptions.ConnectionError as e:
                raise RemoteError(f'Cryptocompare API request failed due to {str(e)}')

            try:
                json_ret = rlk_jsonloads_dict(response.text)
            except JSONDecodeError:
                raise RemoteError(f'Cryptocompare returned invalid JSON response: {response.text}')

            try:
                if json_ret.get('Message', None) == RATE_LIMIT_MSG:
                    if tries >= 1:
                        backoff_seconds = 20 / tries
                        log.debug(
                            f'Got rate limited by cryptocompare. '
                            f'Backing off for {backoff_seconds}',
                        )
                        gevent.sleep(backoff_seconds)
                        tries -= 1
                        continue
                    else:
                        log.debug(
                            f'Got rate limited by cryptocompare and did not manage to get a '
                            f'request through even after {CRYPTOCOMPARE_QUERY_RETRY_TIMES} '
                            f'incremental backoff retries',
                        )

                if json_ret.get('Response', 'Success') != 'Success':
                    error_message = f'Failed to query cryptocompare for: "{querystr}"'
                    if 'Message' in json_ret:
                        error_message += f'. Error: {json_ret["Message"]}'

                    log.error(
                        'Cryptocompare query failure',
                        url=querystr,
                        error=error_message,
                        status_code=response.status_code,
                    )
                    raise RemoteError(error_message)
                return json_ret['Data'] if 'Data' in json_ret else json_ret
            except KeyError as e:
                raise RemoteError(
                    f'Unexpected format of Cryptocompare json_response. '
                    f'Missing key entry for {str(e)}',
                )

        raise AssertionError('We should never get here')

    def _special_case_handling(
            self,
            method_name: Literal[
                'query_endpoint_histohour',
                'query_endpoint_price',
                'query_endpoint_pricehistorical',
            ],
            from_asset: Asset,
            to_asset: Asset,
            **kwargs: Any,
    ) -> Any:
        """Special case handling for queries that need combination of multiple asset queries

        This is hopefully temporary and can be taken care of by cryptocompare itself in the future.

        For some assets cryptocompare can only figure out the price via intermediaries.
        This function takes care of these special cases."""
        method = getattr(self, method_name)
        intermediate_asset = CRYPTOCOMPARE_SPECIAL_CASES_MAPPING[from_asset]
        result1 = method(
            from_asset=from_asset,
            to_asset=intermediate_asset,
            handling_special_case=True,
            **kwargs,
        )
        result2 = method(
            from_asset=intermediate_asset,
            to_asset=to_asset,
            handling_special_case=True,
            **kwargs,
        )
        result: Any
        if method_name == 'query_endpoint_histohour':
            result = {
                'Aggregated': result1['Aggregated'],
                'TimeFrom': result1['TimeFrom'],
                'TimeTo': result1['TimeTo'],
            }
            result1 = result1['Data']
            result2 = result2['Data']
            data = []
            for idx, entry in enumerate(result1):
                entry2 = result2[idx]
                data.append({
                    'time': entry['time'],
                    'high': _multiply_str_nums(entry['high'], entry2['high']),
                    'low': _multiply_str_nums(entry['low'], entry2['low']),
                    'open': _multiply_str_nums(entry['open'], entry2['open']),
                    'volumefrom': entry['volumefrom'],
                    'volumeto': entry['volumeto'],
                    'close': _multiply_str_nums(entry['close'], entry2['close']),
                    'conversionType': entry['conversionType'],
                    'conversionSymbol': entry['conversionSymbol'],
                })
            result['Data'] = data
        elif method_name == 'query_endpoint_price':
            result = {
                to_asset.identifier: _multiply_str_nums(
                    # up until 23/09/2020 cryptocompare may return {} due to bug. Handle
                    # that case by assuming 0 if that happens
                    result1.get(intermediate_asset.identifier, '0'),
                    result2.get(to_asset.identifier, '0'),
                ),
            }
        elif method_name == 'query_endpoint_pricehistorical':
            result = result1 * result2
        else:
            raise RuntimeError(f'Illegal method_name: {method_name}. Should never happen')

        return result

    def query_endpoint_histohour(
            self,
            from_asset: Asset,
            to_asset: Asset,
            limit: int,
            to_timestamp: Timestamp,
            handling_special_case: bool = False,
    ) -> Dict[str, Any]:
        """Returns the full histohour response including TimeFrom and TimeTo

        - May raise RemoteError if there is a problem reaching the cryptocompare server
        or with reading the response returned by the server
        - May raise PriceQueryUnsupportedAsset if from/to assets are not known to cryptocompare
        """
        special_asset = (
            from_asset in CRYPTOCOMPARE_SPECIAL_CASES or to_asset in CRYPTOCOMPARE_SPECIAL_CASES
        )
        if special_asset and not handling_special_case:
            return self._special_case_handling(
                method_name='query_endpoint_histohour',
                from_asset=from_asset,
                to_asset=to_asset,
                limit=limit,
                to_timestamp=to_timestamp,
            )

        try:
            cc_from_asset_symbol = from_asset.to_cryptocompare()
            cc_to_asset_symbol = to_asset.to_cryptocompare()
        except UnsupportedAsset as e:
            raise PriceQueryUnsupportedAsset(e.asset_name)

        query_path = (
            f'v2/histohour?fsym={cc_from_asset_symbol}&tsym={cc_to_asset_symbol}'
            f'&limit={limit}&toTs={to_timestamp}'
        )
        result = self._api_query(path=query_path)
        return result

    def query_endpoint_price(
            self,
            from_asset: Asset,
            to_asset: Asset,
            handling_special_case: bool = False,
    ) -> Dict[str, Any]:
        """Returns the current price of an asset compared to another asset

        - May raise RemoteError if there is a problem reaching the cryptocompare server
        or with reading the response returned by the server
        - May raise PriceQueryUnsupportedAsset if from/to assets are not known to cryptocompare
        """
        special_asset = (
            from_asset in CRYPTOCOMPARE_SPECIAL_CASES or to_asset in CRYPTOCOMPARE_SPECIAL_CASES
        )
        if special_asset and not handling_special_case:
            return self._special_case_handling(
                method_name='query_endpoint_price',
                from_asset=from_asset,
                to_asset=to_asset,
            )
        try:
            cc_from_asset_symbol = from_asset.to_cryptocompare()
            cc_to_asset_symbol = to_asset.to_cryptocompare()
        except UnsupportedAsset as e:
            raise PriceQueryUnsupportedAsset(e.asset_name)

        query_path = f'price?fsym={cc_from_asset_symbol}&tsyms={cc_to_asset_symbol}'
        result = self._api_query(path=query_path)
        return result

    def query_endpoint_pricehistorical(
            self,
            from_asset: Asset,
            to_asset: Asset,
            timestamp: Timestamp,
            handling_special_case: bool = False,
    ) -> Price:
        """Queries the historical daily price of from_asset to to_asset for timestamp

        - May raise RemoteError if there is a problem reaching the cryptocompare server
        or with reading the response returned by the server
        - May raise PriceQueryUnsupportedAsset if from/to assets are not known to cryptocompare
        """
        log.debug(
            'Querying cryptocompare for daily historical price',
            from_asset=from_asset,
            to_asset=to_asset,
            timestamp=timestamp,
        )
        special_asset = (
            from_asset in CRYPTOCOMPARE_SPECIAL_CASES or to_asset in CRYPTOCOMPARE_SPECIAL_CASES
        )
        if special_asset and not handling_special_case:
            return self._special_case_handling(
                method_name='query_endpoint_pricehistorical',
                from_asset=from_asset,
                to_asset=to_asset,
                timestamp=timestamp,
            )

        try:
            cc_from_asset_symbol = from_asset.to_cryptocompare()
            cc_to_asset_symbol = to_asset.to_cryptocompare()
        except UnsupportedAsset as e:
            raise PriceQueryUnsupportedAsset(e.asset_name)

        query_path = (
            f'pricehistorical?fsym={cc_from_asset_symbol}&tsyms={cc_to_asset_symbol}'
            f'&ts={timestamp}'
        )
        if to_asset == 'BTC':
            query_path += '&tryConversion=false'
        result = self._api_query(query_path)
        return Price(FVal(result[cc_from_asset_symbol][cc_to_asset_symbol]))

    def _got_cached_price(self, cache_key: PairCacheKey, timestamp: Timestamp) -> bool:
        """Check if we got a price history for the timestamp cached"""
        if cache_key in self.price_history_file:
            if cache_key not in self.price_history:
                try:
                    with open(self.price_history_file[cache_key], 'r') as f:
                        data = rlk_jsonloads_dict(f.read())
                        self.price_history[cache_key] = _dict_history_to_data(data)
                except (OSError, JSONDecodeError):
                    return False

            in_range = (
                self.price_history[cache_key].start_time <= timestamp and
                self.price_history[cache_key].end_time > timestamp
            )
            if in_range:
                log.debug('Found cached price', cache_key=cache_key, timestamp=timestamp)
                return True

        return False

    def get_historical_data(
            self,
            from_asset: Asset,
            to_asset: Asset,
            timestamp: Timestamp,
            historical_data_start: Timestamp,
    ) -> List[PriceHistoryEntry]:
        """
        Get historical price data from cryptocompare

        Returns a sorted list of price entries.

        - May raise RemoteError if there is a problem reaching the cryptocompare server
        or with reading the response returned by the server
        - May raise UnsupportedAsset if from/to asset is not supported by cryptocompare
        """
        log.debug(
            'Retrieving historical price data from cryptocompare',
            from_asset=from_asset,
            to_asset=to_asset,
            timestamp=timestamp,
        )

        cache_key = PairCacheKey(from_asset.identifier + '_' + to_asset.identifier)
        got_cached_value = self._got_cached_price(cache_key, timestamp)
        if got_cached_value:
            return self.price_history[cache_key].data

        now_ts = ts_now()
        cryptocompare_hourquerylimit = 2000
        calculated_history: List[Dict[str, Any]] = []

        if historical_data_start <= timestamp:
            end_date = historical_data_start
        else:
            end_date = timestamp
        while True:
            pr_end_date = end_date
            end_date = Timestamp(end_date + (cryptocompare_hourquerylimit) * 3600)

            log.debug(
                'Querying cryptocompare for hourly historical price',
                from_asset=from_asset,
                to_asset=to_asset,
                cryptocompare_hourquerylimit=cryptocompare_hourquerylimit,
                end_date=end_date,
            )

            resp = self.query_endpoint_histohour(
                from_asset=from_asset,
                to_asset=to_asset,
                limit=2000,
                to_timestamp=end_date,
            )

            if pr_end_date != resp['TimeFrom']:
                # If we get more than we needed, since we are close to the now_ts
                # then skip all the already included entries
                diff = pr_end_date - resp['TimeFrom']
                # If the start date has less than 3600 secs difference from previous
                # end date then do nothing. If it has more skip all already included entries
                if diff >= 3600:
                    if resp['Data'][diff // 3600]['time'] != pr_end_date:
                        raise RemoteError(
                            'Unexpected fata format in cryptocompare query_endpoint_histohour. '
                            'Expected to find the previous date timestamp during '
                            'cryptocompare historical data fetching',
                        )
                    # just add only the part from the previous timestamp and on
                    resp['Data'] = resp['Data'][diff // 3600:]

            # The end dates of a cryptocompare query do not match. The end date
            # can have up to 3600 secs different to the requested one since this is
            # hourly historical data but no more.
            end_dates_dont_match = (
                end_date < now_ts and
                resp['TimeTo'] != end_date
            )
            if end_dates_dont_match:
                if resp['TimeTo'] - end_date >= 3600:
                    raise RemoteError(
                        'Unexpected fata format in cryptocompare query_endpoint_histohour. '
                        'End dates do not match.',
                    )
                else:
                    # but if it's just a drift within an hour just update the end_date so that
                    # it can be picked up by the next iterations in the loop
                    end_date = resp['TimeTo']

            # If last time slot and first new are the same, skip the first new slot
            last_entry_equal_to_first = (
                len(calculated_history) != 0 and
                calculated_history[-1]['time'] == resp['Data'][0]['time']
            )
            if last_entry_equal_to_first:
                resp['Data'] = resp['Data'][1:]
            calculated_history += resp['Data']
            if end_date >= now_ts:
                break

        # Let's always check for data sanity for the hourly prices.
        _check_hourly_data_sanity(calculated_history, from_asset, to_asset)
        # and now since we actually queried the data let's also cache them
        filename = self.data_directory / ('price_history_' + cache_key + '.json')
        log.info(
            'Updating price history cache',
            filename=filename,
            from_asset=from_asset,
            to_asset=to_asset,
        )
        write_history_data_in_file(
            data=calculated_history,
            filepath=filename,
            start_ts=historical_data_start,
            end_ts=now_ts,
        )

        # Finally save the objects in memory and return them
        data_including_time = {
            'data': calculated_history,
            'start_time': historical_data_start,
            'end_time': end_date,
        }
        self.price_history_file[cache_key] = filename
        self.price_history[cache_key] = _dict_history_to_data(data_including_time)

        return self.price_history[cache_key].data

    def query_historical_price(
            self,
            from_asset: Asset,
            to_asset: Asset,
            timestamp: Timestamp,
            historical_data_start: Timestamp,
    ) -> Price:
        """
        Query the historical price on `timestamp` for `from_asset` in `to_asset`.
        So how much `to_asset` does 1 unit of `from_asset` cost.

        May raise:
        - PriceQueryUnsupportedAsset if from/to asset is known to miss from cryptocompare
        - NoPriceForGivenTimestamp if we can't find a price for the asset in the given
        timestamp from cryptocompare
        - RemoteError if there is a problem reaching the cryptocompare server
        or with reading the response returned by the server
        """

        try:
            data = self.get_historical_data(
                from_asset=from_asset,
                to_asset=to_asset,
                timestamp=timestamp,
                historical_data_start=historical_data_start,
            )
        except UnsupportedAsset as e:
            raise PriceQueryUnsupportedAsset(e.asset_name)

        price = Price(ZERO)
        # all data are sorted and timestamps are always increasing by 1 hour
        # find the closest entry to the provided timestamp
        if timestamp >= data[0].time:
            index_in_bounds = True
            # convert_to_int can't raise here due to its input
            index = convert_to_int((timestamp - data[0].time) / 3600, accept_only_exact=False)
            if index > len(data) - 1:  # index out of bounds
                # Try to see if index - 1 is there and if yes take it
                if index > len(data):
                    index = index - 1
                else:  # give up. This happened: https://github.com/rotki/rotki/issues/1534
                    log.error(
                        f'Expected data index in cryptocompare historical hour price '
                        f'not found. Queried price of: {from_asset.identifier} in '
                        f'{to_asset.identifier} at {timestamp}. Data '
                        f'index: {index}. Length of returned data: {len(data)}. '
                        f'https://github.com/rotki/rotki/issues/1534. Attempting other methods...',
                    )
                    index_in_bounds = False

            if index_in_bounds:
                diff = abs(data[index].time - timestamp)
                if index + 1 <= len(data) - 1:
                    diff_p1 = abs(data[index + 1].time - timestamp)
                    if diff_p1 < diff:
                        index = index + 1

                if data[index].high is not None and data[index].low is not None:
                    price = Price((data[index].high + data[index].low) / 2)

        else:
            # no price found in the historical data from/to asset, try alternatives
            price = Price(ZERO)

        if price == 0:
            if from_asset != 'BTC' and to_asset != 'BTC':
                log.debug(
                    f"Couldn't find historical price from {from_asset} to "
                    f"{to_asset} at timestamp {timestamp}. Comparing with BTC...",
                )
                # Just get the BTC price
                asset_btc_price = PriceHistorian().query_historical_price(
                    from_asset=from_asset,
                    to_asset=Asset('BTC'),
                    timestamp=timestamp,
                )
                btc_to_asset_price = PriceHistorian().query_historical_price(
                    from_asset=Asset('BTC'),
                    to_asset=to_asset,
                    timestamp=timestamp,
                )
                price = Price(asset_btc_price * btc_to_asset_price)
            else:
                log.debug(
                    f"Couldn't find historical price from {from_asset} to "
                    f"{to_asset} at timestamp {timestamp} through cryptocompare."
                    f" Attempting to get daily price...",
                )
                price = self.query_endpoint_pricehistorical(from_asset, to_asset, timestamp)

        comparison_to_nonusd_fiat = (
            (to_asset.is_fiat() and to_asset != Asset('USD')) or
            (from_asset.is_fiat() and from_asset != Asset('USD'))
        )
        if comparison_to_nonusd_fiat:
            price = self._adjust_to_cryptocompare_price_incosistencies(
                price=price,
                from_asset=from_asset,
                to_asset=to_asset,
                timestamp=timestamp,
            )

        if price == 0:
            raise NoPriceForGivenTimestamp(
                from_asset=from_asset,
                to_asset=to_asset,
                date=timestamp_to_date(timestamp, formatstr='%d/%m/%Y, %H:%M:%S'),
            )

        log.debug(
            'Got historical price',
            from_asset=from_asset,
            to_asset=to_asset,
            timestamp=timestamp,
            price=price,
        )

        return price

    @staticmethod
    def _adjust_to_cryptocompare_price_incosistencies(
            price: Price,
            from_asset: Asset,
            to_asset: Asset,
            timestamp: Timestamp,
    ) -> Price:
        """Doublecheck against the USD rate, and if incosistencies are found
        then take the USD adjusted price.

        This is due to incosistencies in the provided historical data from
        cryptocompare. https://github.com/rotki/rotki/issues/221

        Note: Since 12/01/2019 this seems to no longer be happening, but I will
        keep the code around just in case a regression is introduced on the side
        of cryptocompare.

        May raise:
        - PriceQueryUnsupportedAsset if the from asset is known to miss from cryptocompare
        - NoPriceForGivenTimestamp if we can't find a price for the asset in the given
        timestamp from cryptocompare
        - RemoteError if there is a problem reaching the cryptocompare server
        or with reading the response returned by the server
        """
        from_asset_usd = PriceHistorian().query_historical_price(
            from_asset=from_asset,
            to_asset=Asset('USD'),
            timestamp=timestamp,
        )
        to_asset_usd = PriceHistorian().query_historical_price(
            from_asset=to_asset,
            to_asset=Asset('USD'),
            timestamp=timestamp,
        )

        usd_invert_conversion = Price(from_asset_usd / to_asset_usd)
        abs_diff = abs(usd_invert_conversion - price)
        relative_difference = abs_diff / max(price, usd_invert_conversion)
        if relative_difference >= FVal('0.1'):
            log.warning(
                'Cryptocompare historical price data are incosistent.'
                'Taking USD adjusted price. Check github issue #221',
                from_asset=from_asset,
                to_asset=to_asset,
                incosistent_price=price,
                usd_price=from_asset_usd,
                adjusted_price=usd_invert_conversion,
            )
            return usd_invert_conversion
        return price

    def all_coins(self) -> Dict[str, Any]:
        """
        Gets the list of all the cryptocompare coins

        May raise:
        - RemoteError if there is a problem reaching the cryptocompare server
        or with reading the response returned by the server
        """
        # Get coin list of cryptocompare
        invalidate_cache = True
        coinlist_cache_path = os.path.join(self.data_directory, 'cryptocompare_coinlist.json')
        if os.path.isfile(coinlist_cache_path):
            log.info('Found cryptocompare coinlist cache', path=coinlist_cache_path)
            with open(coinlist_cache_path, 'r') as f:
                try:
                    data = rlk_jsonloads_dict(f.read())
                    now = ts_now()
                    invalidate_cache = False

                    # If we got a cache and its' over a month old then requery cryptocompare
                    if data['time'] < now and now - data['time'] > 2629800:
                        log.info('Cryptocompare coinlist cache is now invalidated')
                        invalidate_cache = True
                        data = data['data']
                except JSONDecodeError:
                    invalidate_cache = True

        if invalidate_cache:
            data = self._api_query('all/coinlist')

            # Also save the cache
            with open(coinlist_cache_path, 'w') as f:
                now = ts_now()
                log.info('Writing coinlist cache', timestamp=now)
                write_data = {'time': now, 'data': data}
                f.write(rlk_jsondumps(write_data))
        else:
            # in any case take the data
            data = data['data']

        # As described in the docs
        # https://min-api.cryptocompare.com/documentation?key=Other&cat=allCoinsWithContentEndpoint
        # This is not the entire list of assets in the system, so I am manually adding
        # here assets I am aware of that they already have historical data for in thei
        # cryptocompare system
        data['DAO'] = object()
        data['USDT'] = object()
        data['VEN'] = object()
        data['AIR*'] = object()  # This is Aircoin
        # This is SpendCoin (https://coinmarketcap.com/currencies/spendcoin/)
        data['SPND'] = object()
        # This is eBitcoinCash (https://coinmarketcap.com/currencies/ebitcoin-cash/)
        data['EBCH'] = object()
        # This is Educare (https://coinmarketcap.com/currencies/educare/)
        data['EKT'] = object()
        # This is Knoxstertoken (https://coinmarketcap.com/currencies/knoxstertoken/)
        data['FKX'] = object()
        # This is FNKOS (https://coinmarketcap.com/currencies/fnkos/)
        data['FNKOS'] = object()
        # This is FansTime (https://coinmarketcap.com/currencies/fanstime/)
        data['FTI'] = object()
        # This is Gene Source Code Chain
        # (https://coinmarketcap.com/currencies/gene-source-code-chain/)
        data['GENE*'] = object()
        # This is GazeCoin (https://coinmarketcap.com/currencies/gazecoin/)
        data['GZE'] = object()
        # This is probaly HarmonyCoin (https://coinmarketcap.com/currencies/harmonycoin-hmc/)
        data['HMC*'] = object()
        # This is IoTChain (https://coinmarketcap.com/currencies/iot-chain/)
        data['ITC'] = object()
        # This is Luna Coin (https://coinmarketcap.com/currencies/luna-coin/)
        data['LUNA'] = object
        # This is MFTU (https://coinmarketcap.com/currencies/mainstream-for-the-underground/)
        data['MFTU'] = object()
        # This is Nexxus (https://coinmarketcap.com/currencies/nexxus/)
        data['NXX'] = object()
        # This is Owndata (https://coinmarketcap.com/currencies/owndata/)
        data['OWN'] = object()
        # This is PiplCoin (https://coinmarketcap.com/currencies/piplcoin/)
        data['PIPL'] = object()
        # This is PKG Token (https://coinmarketcap.com/currencies/pkg-token/)
        data['PKG'] = object()
        # This is Quibitica https://coinmarketcap.com/currencies/qubitica/
        data['QBIT'] = object()
        # This is DPRating https://coinmarketcap.com/currencies/dprating/
        data['RATING'] = object()
        # This is RouletteToken https://coinmarketcap.com/currencies/roulettetoken/
        data['RLT'] = object()
        # This is RocketPool https://coinmarketcap.com/currencies/rocket-pool/
        data['RPL'] = object()
        # This is SpeedMiningService (https://coinmarketcap.com/currencies/speed-mining-service/)
        data['SMS'] = object()
        # This is SmartShare (https://coinmarketcap.com/currencies/smartshare/)
        data['SSP'] = object()
        # This is ThoreCoin (https://coinmarketcap.com/currencies/thorecoin/)
        data['THR'] = object()
        # This is Transcodium (https://coinmarketcap.com/currencies/transcodium/)
        data['TNS'] = object()

        return data
